from dataclasses import replace
from pathlib import Path
from typing import Any, cast

import anyio
import pytest

from yee88 import commands, plugins
from yee88.telegram.commands.executor import _CaptureTransport, _run_engine
from yee88.telegram.commands.file_transfer import _handle_file_get, _handle_file_put
from yee88.telegram.commands.model import _handle_model_command
from yee88.telegram.commands.reasoning import _handle_reasoning_command
from yee88.telegram.commands.topics import _handle_topic_command
import yee88.telegram.loop as telegram_loop
import yee88.telegram.topics as telegram_topics
from yee88.directives import parse_directives
from yee88.telegram.api_models import Chat, File, ForumTopic, Message, Update, User
from yee88.settings import TelegramFilesSettings, TelegramTopicsSettings
from yee88.telegram.bridge import (
    TelegramBridgeConfig,
    TelegramPresenter,
    TelegramTransport,
    build_bot_commands,
    handle_callback_cancel,
    handle_cancel,
    is_cancel_command,
    run_main_loop,
    send_with_resume,
)
from yee88.telegram.client import BotClient
from yee88.telegram.render import MAX_BODY_CHARS
from yee88.telegram.topic_state import TopicStateStore, resolve_state_path
from yee88.telegram.chat_sessions import ChatSessionStore, resolve_sessions_path
from yee88.telegram.chat_prefs import ChatPrefsStore, resolve_prefs_path
from yee88.telegram.engine_overrides import EngineOverrides
from yee88.context import RunContext
from yee88.config import ProjectConfig, ProjectsConfig
from yee88.runner_bridge import ExecBridgeConfig, RunningTask
from yee88.markdown import MarkdownPresenter
from yee88.model import ResumeToken
from yee88.progress import ProgressTracker
from yee88.router import AutoRouter, RunnerEntry
from yee88.scheduler import ThreadScheduler
from yee88.transport_runtime import TransportRuntime
from yee88.runners.mock import Return, ScriptRunner, Sleep, Wait
from yee88.telegram.types import (
    TelegramCallbackQuery,
    TelegramDocument,
    TelegramIncomingMessage,
    TelegramVoice,
)
from yee88.transport import MessageRef, RenderedMessage, SendOptions
from tests.plugin_fixtures import FakeEntryPoint, install_entrypoints
from tests.telegram_fakes import (
    FakeBot,
    FakeTransport,
    _empty_projects,
    make_cfg,
    _make_router,
)

CODEX_ENGINE = "codex"
FAST_FORWARD_COALESCE_S = 0.0
FAST_MEDIA_GROUP_DEBOUNCE_S = 0.0
BATCH_MEDIA_GROUP_DEBOUNCE_S = 0.05
DEBOUNCE_FORWARD_COALESCE_S = 0.05


class _NoopTaskGroup:
    def start_soon(self, func, *args: Any) -> None:
        _ = func, args
        return None


def test_parse_directives_inline_engine() -> None:
    directives = parse_directives(
        "/claude do it",
        engine_ids=("codex", "claude"),
        projects=_empty_projects(),
    )
    assert directives.engine == "claude"
    assert directives.prompt == "do it"


def test_parse_directives_newline() -> None:
    directives = parse_directives(
        "/codex\nhello",
        engine_ids=("codex", "claude"),
        projects=_empty_projects(),
    )
    assert directives.engine == "codex"
    assert directives.prompt == "hello"


def test_parse_directives_ignores_unknown() -> None:
    directives = parse_directives(
        "/unknown hi",
        engine_ids=("codex", "claude"),
        projects=_empty_projects(),
    )
    assert directives.engine is None
    assert directives.prompt == "/unknown hi"


def test_parse_directives_bot_suffix() -> None:
    directives = parse_directives(
        "/claude@bunny_agent_bot hi",
        engine_ids=("claude",),
        projects=_empty_projects(),
    )
    assert directives.engine == "claude"
    assert directives.prompt == "hi"


def test_parse_directives_only_first_non_empty_line() -> None:
    directives = parse_directives(
        "hello\n/claude hi",
        engine_ids=("codex", "claude"),
        projects=_empty_projects(),
    )
    assert directives.engine is None
    assert directives.prompt == "hello\n/claude hi"


def test_build_bot_commands_includes_cancel_and_engine() -> None:
    runner = ScriptRunner(
        [Return(answer="ok")], engine=CODEX_ENGINE, resume_value="sid"
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )
    commands = build_bot_commands(runtime)

    assert {"command": "cancel", "description": "cancel run"} in commands
    assert {"command": "file", "description": "upload or fetch files"} in commands
    assert {"command": "new", "description": "start a new thread"} in commands
    assert {"command": "ctx", "description": "show or update context"} in commands
    assert {"command": "agent", "description": "set default engine"} in commands
    assert any(cmd["command"] == "codex" for cmd in commands)


def test_build_bot_commands_includes_projects() -> None:
    runner = ScriptRunner(
        [Return(answer="ok")], engine=CODEX_ENGINE, resume_value="sid"
    )
    router = _make_router(runner)
    projects = ProjectsConfig(
        projects={
            "good": ProjectConfig(
                alias="good",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
            ),
            "bad-name": ProjectConfig(
                alias="bad-name",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
            ),
        },
        default_project=None,
    )

    runtime = TransportRuntime(router=router, projects=projects)
    commands = build_bot_commands(runtime)

    assert any(cmd["command"] == "good" for cmd in commands)
    assert not any(cmd["command"] == "bad-name" for cmd in commands)


def test_build_bot_commands_includes_topics_when_enabled() -> None:
    runner = ScriptRunner(
        [Return(answer="ok")], engine=CODEX_ENGINE, resume_value="sid"
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )

    commands = build_bot_commands(runtime, include_topics=True)

    assert {"command": "topic", "description": "create or bind a topic"} in commands
    assert {"command": "ctx", "description": "show or update context"} in commands


def test_build_bot_commands_includes_command_plugins(monkeypatch) -> None:
    class _Command:
        id = "pingcmd"
        description = "ping command"

        async def handle(self, ctx):
            _ = ctx
            return None

    entrypoints = [
        FakeEntryPoint(
            "pingcmd",
            "yee88.commands.ping:BACKEND",
            plugins.COMMAND_GROUP,
            loader=_Command,
        )
    ]
    install_entrypoints(monkeypatch, entrypoints)
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )

    commands_list = build_bot_commands(runtime)

    assert {"command": "pingcmd", "description": "ping command"} in commands_list


def test_build_bot_commands_caps_total() -> None:
    runner = ScriptRunner(
        [Return(answer="ok")], engine=CODEX_ENGINE, resume_value="sid"
    )
    router = _make_router(runner)
    projects = ProjectsConfig(
        projects={
            f"proj{i}": ProjectConfig(
                alias=f"proj{i}",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
            )
            for i in range(150)
        },
        default_project=None,
    )

    runtime = TransportRuntime(router=router, projects=projects)
    commands = build_bot_commands(runtime)

    assert len(commands) == 100
    assert any(cmd["command"] == "codex" for cmd in commands)
    assert any(cmd["command"] == "cancel" for cmd in commands)


def test_telegram_presenter_progress_shows_cancel_button() -> None:
    presenter = TelegramPresenter()
    state = ProgressTracker(engine="codex").snapshot()

    rendered = presenter.render_progress(state, elapsed_s=0.0)

    reply_markup = rendered.extra["reply_markup"]
    assert reply_markup["inline_keyboard"][0][0]["text"] == "cancel"
    assert reply_markup["inline_keyboard"][0][0]["callback_data"] == "yee88:cancel"


def test_telegram_presenter_clears_button_on_cancelled() -> None:
    presenter = TelegramPresenter()
    state = ProgressTracker(engine="codex").snapshot()

    rendered = presenter.render_progress(state, elapsed_s=0.0, label="`cancelled`")

    assert rendered.extra["reply_markup"]["inline_keyboard"] == []


def test_telegram_presenter_final_clears_button() -> None:
    presenter = TelegramPresenter()
    state = ProgressTracker(engine="codex").snapshot()

    rendered = presenter.render_final(state, elapsed_s=0.0, status="done", answer="ok")

    assert rendered.extra["reply_markup"]["inline_keyboard"] == []


def test_telegram_presenter_split_overflow_adds_followups() -> None:
    presenter = TelegramPresenter(message_overflow="split")
    state = ProgressTracker(engine="codex").snapshot()

    rendered = presenter.render_final(
        state,
        elapsed_s=0.0,
        status="done",
        answer="x" * (MAX_BODY_CHARS + 10),
    )

    followups = rendered.extra.get("followups")
    assert followups
    assert all(isinstance(item, RenderedMessage) for item in followups)
    assert rendered.extra["reply_markup"]["inline_keyboard"] == []
    assert all(
        item.extra["reply_markup"]["inline_keyboard"] == [] for item in followups
    )


@pytest.mark.anyio
async def test_telegram_transport_passes_replace_and_wait() -> None:
    bot = FakeBot()
    transport = TelegramTransport(bot)
    reply = MessageRef(channel_id=123, message_id=10)
    replace = MessageRef(channel_id=123, message_id=11)

    await transport.send(
        channel_id=123,
        message=RenderedMessage(text="hello"),
        options=SendOptions(reply_to=reply, notify=True, replace=replace),
    )
    assert bot.send_calls
    assert bot.send_calls[0]["replace_message_id"] == 11

    await transport.edit(
        ref=replace,
        message=RenderedMessage(text="edit"),
        wait=False,
    )
    assert bot.edit_calls
    assert bot.edit_calls[0]["wait"] is False


@pytest.mark.anyio
async def test_telegram_transport_passes_reply_markup() -> None:
    bot = FakeBot()
    transport = TelegramTransport(bot)
    markup = {"inline_keyboard": []}

    await transport.send(
        channel_id=123,
        message=RenderedMessage(text="hello", extra={"reply_markup": markup}),
    )
    assert bot.send_calls
    assert bot.send_calls[0]["reply_markup"] == markup

    ref = MessageRef(channel_id=123, message_id=1)
    await transport.edit(
        ref=ref,
        message=RenderedMessage(text="edit", extra={"reply_markup": markup}),
    )
    assert bot.edit_calls
    assert bot.edit_calls[0]["reply_markup"] == markup


@pytest.mark.anyio
async def test_telegram_transport_sends_followups() -> None:
    bot = FakeBot()
    transport = TelegramTransport(bot)
    reply = MessageRef(channel_id=123, message_id=10)
    followup = RenderedMessage(text="part 2")

    await transport.send(
        channel_id=123,
        message=RenderedMessage(text="part 1", extra={"followups": [followup]}),
        options=SendOptions(reply_to=reply, notify=False, thread_id=7),
    )

    assert len(bot.send_calls) == 2
    assert bot.send_calls[1]["text"] == "part 2"
    assert bot.send_calls[1]["reply_to_message_id"] == 10
    assert bot.send_calls[1]["message_thread_id"] == 7
    assert bot.send_calls[1]["replace_message_id"] is None
    assert bot.send_calls[1]["disable_notification"] is True


@pytest.mark.anyio
async def test_telegram_transport_edits_and_sends_followups() -> None:
    bot = FakeBot()
    transport = TelegramTransport(bot)
    followup = RenderedMessage(text="part 2")

    await transport.edit(
        ref=MessageRef(channel_id=123, message_id=42),
        message=RenderedMessage(
            text="part 1",
            extra={
                "followups": [followup],
                "followup_reply_to_message_id": 10,
                "followup_thread_id": 7,
                "followup_notify": False,
            },
        ),
    )

    assert len(bot.edit_calls) == 1
    assert len(bot.send_calls) == 1
    assert bot.send_calls[0]["text"] == "part 2"
    assert bot.send_calls[0]["reply_to_message_id"] == 10
    assert bot.send_calls[0]["message_thread_id"] == 7
    assert bot.send_calls[0]["disable_notification"] is True


@pytest.mark.anyio
async def test_telegram_transport_edit_wait_false_returns_ref() -> None:
    class _OutboxBot(BotClient):
        def __init__(self) -> None:
            self.edit_calls: list[dict[str, Any]] = []

        async def get_updates(
            self,
            offset: int | None,
            timeout_s: int = 50,
            allowed_updates: list[str] | None = None,
        ) -> list[Update] | None:
            return None

        async def get_file(self, file_id: str) -> File | None:
            _ = file_id
            return None

        async def download_file(self, file_path: str) -> bytes | None:
            _ = file_path
            return None

        async def send_message(
            self,
            chat_id: int,
            text: str,
            reply_to_message_id: int | None = None,
            disable_notification: bool | None = False,
            message_thread_id: int | None = None,
            entities: list[dict[str, Any]] | None = None,
            parse_mode: str | None = None,
            reply_markup: dict | None = None,
            *,
            replace_message_id: int | None = None,
        ) -> Message | None:
            _ = reply_markup
            return None

        async def send_document(
            self,
            chat_id: int,
            filename: str,
            content: bytes,
            reply_to_message_id: int | None = None,
            message_thread_id: int | None = None,
            disable_notification: bool | None = False,
            caption: str | None = None,
        ) -> Message | None:
            _ = (
                chat_id,
                filename,
                content,
                reply_to_message_id,
                message_thread_id,
                disable_notification,
                caption,
            )
            return None

        async def edit_message_text(
            self,
            chat_id: int,
            message_id: int,
            text: str,
            entities: list[dict[str, Any]] | None = None,
            parse_mode: str | None = None,
            reply_markup: dict | None = None,
            *,
            wait: bool = True,
        ) -> Message | None:
            self.edit_calls.append(
                {
                    "chat_id": chat_id,
                    "message_id": message_id,
                    "text": text,
                    "entities": entities,
                    "parse_mode": parse_mode,
                    "reply_markup": reply_markup,
                    "wait": wait,
                }
            )
            if not wait:
                return None
            return Message(message_id=message_id, chat=Chat(id=chat_id, type="private"))

        async def delete_message(
            self,
            chat_id: int,
            message_id: int,
        ) -> bool:
            return False

        async def set_my_commands(
            self,
            commands: list[dict[str, Any]],
            *,
            scope: dict[str, Any] | None = None,
            language_code: str | None = None,
        ) -> bool:
            return False

        async def get_me(self) -> User | None:
            return None

        async def close(self) -> None:
            return None

        async def answer_callback_query(
            self,
            callback_query_id: str,
            text: str | None = None,
            show_alert: bool | None = None,
        ) -> bool:
            _ = callback_query_id, text, show_alert
            return True

    bot = _OutboxBot()
    transport = TelegramTransport(bot)
    ref = MessageRef(channel_id=123, message_id=1)

    result = await transport.edit(
        ref=ref,
        message=RenderedMessage(text="edit"),
        wait=False,
    )

    assert result == ref
    assert bot.edit_calls
    assert bot.edit_calls[0]["wait"] is False


@pytest.mark.anyio
async def test_handle_cancel_without_reply_prompts_user() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/cancel",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=123,
    )
    running_tasks: dict = {}

    await handle_cancel(cfg, msg, running_tasks)

    assert len(transport.send_calls) == 1
    assert "reply to the progress message" in transport.send_calls[0]["message"].text


@pytest.mark.anyio
async def test_handle_cancel_with_no_progress_message_says_nothing_running() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/cancel",
        reply_to_message_id=None,
        reply_to_text="no message id",
        sender_id=123,
    )
    running_tasks: dict = {}

    await handle_cancel(cfg, msg, running_tasks)

    assert len(transport.send_calls) == 1
    assert "nothing is currently running" in transport.send_calls[0]["message"].text


@pytest.mark.anyio
async def test_handle_cancel_with_finished_task_says_nothing_running() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    progress_id = 99
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/cancel",
        reply_to_message_id=progress_id,
        reply_to_text=None,
        sender_id=123,
    )
    running_tasks: dict = {}

    await handle_cancel(cfg, msg, running_tasks)

    assert len(transport.send_calls) == 1
    assert "nothing is currently running" in transport.send_calls[0]["message"].text


@pytest.mark.anyio
async def test_handle_cancel_cancels_running_task() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    progress_id = 42
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/cancel",
        reply_to_message_id=progress_id,
        reply_to_text=None,
        sender_id=123,
    )

    running_task = RunningTask()
    running_tasks = {MessageRef(channel_id=123, message_id=progress_id): running_task}
    await handle_cancel(cfg, msg, running_tasks)

    assert running_task.cancel_requested.is_set() is True
    assert len(transport.send_calls) == 0  # No error message sent


@pytest.mark.anyio
async def test_handle_cancel_only_cancels_matching_progress_message() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    task_first = RunningTask()
    task_second = RunningTask()
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/cancel",
        reply_to_message_id=1,
        reply_to_text=None,
        sender_id=123,
    )
    running_tasks = {
        MessageRef(channel_id=123, message_id=1): task_first,
        MessageRef(channel_id=123, message_id=2): task_second,
    }

    await handle_cancel(cfg, msg, running_tasks)

    assert task_first.cancel_requested.is_set() is True
    assert task_second.cancel_requested.is_set() is False
    assert len(transport.send_calls) == 0


@pytest.mark.anyio
async def test_handle_cancel_cancels_queued_job() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)

    async def _noop_run_job(_) -> None:
        return None

    scheduler = ThreadScheduler(task_group=_NoopTaskGroup(), run_job=_noop_run_job)
    progress_id = 55
    progress_ref = MessageRef(channel_id=123, message_id=progress_id)
    resume = ResumeToken(engine=CODEX_ENGINE, value="sid")
    await scheduler.enqueue_resume(
        chat_id=123,
        user_msg_id=10,
        text="queued",
        resume_token=resume,
        progress_ref=progress_ref,
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/cancel",
        reply_to_message_id=progress_id,
        reply_to_text=None,
        sender_id=123,
    )

    await handle_cancel(cfg, msg, {}, scheduler)

    assert transport.edit_calls
    assert "cancelled" in transport.edit_calls[0]["message"].text.lower()
    assert await scheduler.cancel_queued(123, progress_ref.message_id) is None


@pytest.mark.anyio
async def test_handle_file_put_writes_file(tmp_path: Path) -> None:
    payload = b"hello"

    class _FileBot(FakeBot):
        async def get_file(self, file_id: str) -> File | None:
            _ = file_id
            return File(file_path="files/hello.txt")

        async def download_file(self, file_path: str) -> bytes | None:
            _ = file_path
            return payload

    transport = FakeTransport()
    bot = _FileBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project=None,
    )
    runtime = TransportRuntime(router=_make_router(runner), projects=projects)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        files=TelegramFilesSettings(enabled=True),
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=321,
        chat_type="private",
        document=TelegramDocument(
            file_id="doc-id",
            file_name="hello.txt",
            mime_type="text/plain",
            file_size=len(payload),
            raw={"file_id": "doc-id"},
        ),
    )

    await _handle_file_put(cfg, msg, "/proj uploads/hello.txt", None, None)

    target = tmp_path / "uploads" / "hello.txt"
    assert target.read_bytes() == payload
    assert transport.send_calls
    text = transport.send_calls[-1]["message"].text
    assert "saved uploads/hello.txt" in text
    assert "(5 b)" in text


@pytest.mark.anyio
async def test_handle_file_get_sends_document_for_allowed_user(
    tmp_path: Path,
) -> None:
    payload = b"fetch"
    target = tmp_path / "hello.txt"
    target.write_bytes(payload)

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project=None,
    )
    runtime = TransportRuntime(router=_make_router(runner), projects=projects)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        files=TelegramFilesSettings(
            enabled=True,
            allowed_user_ids=[42],
        ),
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=-100,
        message_id=10,
        text="",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=42,
        chat_type="supergroup",
    )

    await _handle_file_get(cfg, msg, "/proj hello.txt", None, None)

    assert bot.document_calls
    assert bot.document_calls[0]["filename"] == "hello.txt"
    assert bot.document_calls[0]["content"] == payload


@pytest.mark.anyio
async def test_handle_callback_cancel_cancels_running_task() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    progress_id = 42
    running_task = RunningTask()
    running_tasks = {MessageRef(channel_id=123, message_id=progress_id): running_task}
    query = TelegramCallbackQuery(
        transport="telegram",
        chat_id=123,
        message_id=progress_id,
        callback_query_id="cbq-1",
        data="yee88:cancel",
        sender_id=123,
    )

    await handle_callback_cancel(cfg, query, running_tasks)

    assert running_task.cancel_requested.is_set() is True
    assert len(transport.send_calls) == 0
    bot = cast(FakeBot, cfg.bot)
    assert bot.callback_calls
    assert bot.callback_calls[-1]["text"] == "cancelling..."


@pytest.mark.anyio
async def test_handle_callback_cancel_cancels_queued_job() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)

    async def _noop_run_job(_) -> None:
        return None

    scheduler = ThreadScheduler(task_group=_NoopTaskGroup(), run_job=_noop_run_job)
    progress_id = 77
    progress_ref = MessageRef(channel_id=123, message_id=progress_id)
    resume = ResumeToken(engine=CODEX_ENGINE, value="sid")
    await scheduler.enqueue_resume(
        chat_id=123,
        user_msg_id=10,
        text="queued",
        resume_token=resume,
        progress_ref=progress_ref,
    )
    query = TelegramCallbackQuery(
        transport="telegram",
        chat_id=123,
        message_id=progress_id,
        callback_query_id="cbq-queued",
        data="yee88:cancel",
        sender_id=123,
    )

    await handle_callback_cancel(cfg, query, {}, scheduler)

    assert transport.edit_calls
    assert "cancelled" in transport.edit_calls[0]["message"].text.lower()
    bot = cast(FakeBot, cfg.bot)
    assert bot.callback_calls
    assert bot.callback_calls[-1]["text"] == "dropped from queue."


@pytest.mark.anyio
async def test_handle_callback_cancel_without_task_acknowledges() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    query = TelegramCallbackQuery(
        transport="telegram",
        chat_id=123,
        message_id=99,
        callback_query_id="cbq-2",
        data="yee88:cancel",
        sender_id=123,
    )

    await handle_callback_cancel(cfg, query, {})

    assert len(transport.send_calls) == 0
    bot = cast(FakeBot, cfg.bot)
    assert bot.callback_calls
    assert "nothing is currently running" in bot.callback_calls[-1]["text"].lower()


def test_allowed_chat_ids_include_allowed_user_ids() -> None:
    cfg = replace(make_cfg(FakeTransport()), allowed_user_ids=(42,))
    allowed = telegram_loop._allowed_chat_ids(cfg)
    assert cfg.chat_id in allowed
    assert 42 in allowed


@pytest.mark.anyio
async def test_run_main_loop_ignores_disallowed_sender() -> None:
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    cfg = replace(make_cfg(FakeTransport(), runner), allowed_user_ids=(999,))

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert runner.calls == []


@pytest.mark.anyio
async def test_run_main_loop_ignores_disallowed_callback() -> None:
    cfg = replace(make_cfg(FakeTransport()), allowed_user_ids=(999,))
    bot = cast(FakeBot, cfg.bot)

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramCallbackQuery(
            transport="telegram",
            chat_id=123,
            message_id=42,
            callback_query_id="cbq-ignored",
            data="yee88:cancel",
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert bot.callback_calls == []


@pytest.mark.anyio
async def test_run_main_loop_allows_allowed_sender() -> None:
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    cfg = replace(make_cfg(FakeTransport(), runner), allowed_user_ids=(123,))

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert runner.calls
    assert runner.calls[0][0] == "hello"


def test_cancel_command_accepts_extra_text() -> None:
    assert is_cancel_command("/cancel now") is True
    assert is_cancel_command("/cancel@yee88 please") is True
    assert is_cancel_command("/cancelled") is False


def test_resolve_message_accepts_backticked_ctx_line() -> None:
    runtime = TransportRuntime(
        router=_make_router(ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)),
        projects=ProjectsConfig(
            projects={
                "yee88": ProjectConfig(
                    alias="yee88",
                    path=Path("."),
                    worktrees_dir=Path(".worktrees"),
                )
            },
            default_project=None,
        ),
    )
    resolved = runtime.resolve_message(
        text="do it",
        reply_text="`ctx: yee88 @feat/api`",
    )

    assert resolved.prompt == "do it"
    assert resolved.resume_token is None
    assert resolved.engine_override is None
    assert resolved.context == RunContext(project="yee88", branch="feat/api")


def test_is_forwarded_detects_forward_fields() -> None:
    assert telegram_loop._is_forwarded({"forward_origin": {"type": "user"}})
    assert telegram_loop._is_forwarded({"forward_from": {"id": 1}})
    assert telegram_loop._is_forwarded({"forward_from_chat": {"id": 1}})
    assert telegram_loop._is_forwarded({"forward_from_message_id": 2})
    assert telegram_loop._is_forwarded({"forward_sender_name": "anon"})
    assert telegram_loop._is_forwarded({"forward_signature": "sig"})
    assert telegram_loop._is_forwarded({"forward_date": 123})
    assert telegram_loop._is_forwarded({"is_automatic_forward": True})
    assert not telegram_loop._is_forwarded({"text": "hello"})
    assert not telegram_loop._is_forwarded(None)


def test_topic_title_matches_command_syntax() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)

    title = telegram_topics._topic_title(
        runtime=cfg.runtime,
        context=RunContext(project="yee88", branch="master"),
    )

    assert title == "yee88 @master"

    title = telegram_topics._topic_title(
        runtime=cfg.runtime,
        context=RunContext(project="yee88", branch=None),
    )

    assert title == "yee88"

    title = telegram_topics._topic_title(
        runtime=cfg.runtime,
        context=RunContext(project=None, branch="main"),
    )

    assert title == "@main"


def test_topic_title_projects_scope_includes_project() -> None:
    transport = FakeTransport()
    cfg = replace(
        make_cfg(transport),
        topics=TelegramTopicsSettings(
            enabled=True,
            scope="projects",
        ),
    )

    title = telegram_topics._topic_title(
        runtime=cfg.runtime,
        context=RunContext(project="yee88", branch="master"),
    )

    assert title == "yee88 @master"


@pytest.mark.anyio
async def test_maybe_rename_topic_updates_title(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    store = TopicStateStore(tmp_path / "telegram_topics_state.json")

    await store.set_context(
        123,
        77,
        RunContext(project="yee88", branch="old"),
        topic_title="yee88 @old",
    )

    await telegram_topics._maybe_rename_topic(
        cfg,
        store,
        chat_id=123,
        thread_id=77,
        context=RunContext(project="yee88", branch="new"),
    )

    bot = cast(FakeBot, cfg.bot)
    assert bot.edit_topic_calls
    assert bot.edit_topic_calls[-1]["name"] == "yee88 @new"
    snapshot = await store.get_thread(123, 77)
    assert snapshot is not None
    assert snapshot.topic_title == "yee88 @new"


@pytest.mark.anyio
async def test_maybe_rename_topic_skips_when_title_matches(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    store = TopicStateStore(tmp_path / "telegram_topics_state.json")

    await store.set_context(
        123,
        77,
        RunContext(project="yee88", branch="main"),
        topic_title="yee88 @main",
    )
    snapshot = await store.get_thread(123, 77)

    await telegram_topics._maybe_rename_topic(
        cfg,
        store,
        chat_id=123,
        thread_id=77,
        context=RunContext(project="yee88", branch="main"),
        snapshot=snapshot,
    )

    bot = cast(FakeBot, cfg.bot)
    assert bot.edit_topic_calls == []


@pytest.mark.anyio
async def test_topic_command_recreates_stale_topic(tmp_path: Path) -> None:
    class _StaleTopicBot(FakeBot):
        def __init__(self) -> None:
            super().__init__()
            self.create_topic_calls: list[dict[str, Any]] = []

        async def create_forum_topic(
            self, chat_id: int, name: str
        ) -> ForumTopic | None:
            self.create_topic_calls.append({"chat_id": chat_id, "name": name})
            return ForumTopic(message_thread_id=55)

        async def edit_forum_topic(
            self, chat_id: int, message_thread_id: int, name: str
        ) -> bool:
            self.edit_topic_calls.append(
                {
                    "chat_id": chat_id,
                    "message_thread_id": message_thread_id,
                    "name": name,
                }
            )
            return False

    transport = FakeTransport()
    bot = _StaleTopicBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    projects = ProjectsConfig(
        projects={
            "yee88": ProjectConfig(
                alias="yee88",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project=None,
    )
    runtime = TransportRuntime(router=_make_router(runner), projects=projects)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        topics=TelegramTopicsSettings(enabled=True, scope="main"),
    )
    store = TopicStateStore(tmp_path / "telegram_topics_state.json")
    await store.set_context(
        123,
        77,
        RunContext(project="yee88", branch="master"),
        topic_title="yee88 @master",
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/topic yee88 @master",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=123,
    )

    await _handle_topic_command(
        cfg,
        msg,
        "yee88 @master",
        store,
        resolved_scope="main",
        scope_chat_ids=frozenset({123}),
    )

    assert bot.edit_topic_calls
    assert bot.create_topic_calls
    assert await store.get_thread(123, 77) is None
    snapshot = await store.get_thread(123, 55)
    assert snapshot is not None
    assert snapshot.context == RunContext(project="yee88", branch="master")


@pytest.mark.anyio
async def test_model_command_show_reports_overrides(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    cfg = replace(cfg, topics=TelegramTopicsSettings(enabled=True, scope="main"))
    chat_prefs = ChatPrefsStore(tmp_path / "telegram_chat_prefs_state.json")
    topic_store = TopicStateStore(tmp_path / "telegram_topics_state.json")
    await chat_prefs.set_engine_override(
        123,
        CODEX_ENGINE,
        EngineOverrides(model="gpt-4.1-mini", reasoning=None),
    )
    await topic_store.set_engine_override(
        123,
        77,
        CODEX_ENGINE,
        EngineOverrides(model="gpt-4.1", reasoning=None),
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/model",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=123,
        thread_id=77,
    )

    await _handle_model_command(
        cfg,
        msg,
        "",
        ambient_context=None,
        topic_store=topic_store,
        chat_prefs=chat_prefs,
        resolved_scope="main",
        scope_chat_ids=frozenset({123}),
    )

    text = transport.send_calls[-1]["message"].text
    assert "engine: codex (global default)" in text
    assert "model: gpt-4.1 (topic override)" in text
    assert "defaults: topic: gpt-4.1, chat: gpt-4.1-mini" in text
    assert "available engines: codex" in text


@pytest.mark.anyio
async def test_model_command_set_and_clear_chat_override(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    chat_prefs = ChatPrefsStore(tmp_path / "telegram_chat_prefs_state.json")
    await chat_prefs.set_engine_override(
        123,
        CODEX_ENGINE,
        EngineOverrides(model=None, reasoning="low"),
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/model set gpt-4.1-mini",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=456,
        chat_type="supergroup",
    )

    await _handle_model_command(
        cfg,
        msg,
        "set gpt-4.1-mini",
        ambient_context=None,
        topic_store=None,
        chat_prefs=chat_prefs,
    )

    override = await chat_prefs.get_engine_override(123, CODEX_ENGINE)
    assert override is not None
    assert override.model == "gpt-4.1-mini"
    assert override.reasoning == "low"
    assert (
        "chat model override set to gpt-4.1-mini for codex."
        in transport.send_calls[-1]["message"].text
    )

    msg_clear = replace(
        msg,
        message_id=11,
        text="/model clear codex",
    )
    await _handle_model_command(
        cfg,
        msg_clear,
        "clear codex",
        ambient_context=None,
        topic_store=None,
        chat_prefs=chat_prefs,
    )

    override = await chat_prefs.get_engine_override(123, CODEX_ENGINE)
    assert override is not None
    assert override.model is None
    assert override.reasoning == "low"
    assert "chat model override cleared." in transport.send_calls[-1]["message"].text


@pytest.mark.anyio
async def test_reasoning_command_set_and_clear_topic_override(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    cfg = replace(cfg, topics=TelegramTopicsSettings(enabled=True, scope="main"))
    topic_store = TopicStateStore(tmp_path / "telegram_topics_state.json")
    await topic_store.set_engine_override(
        123,
        77,
        CODEX_ENGINE,
        EngineOverrides(model="gpt-4.1-mini", reasoning=None),
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/reasoning set High",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=456,
        chat_type="supergroup",
        thread_id=77,
    )

    await _handle_reasoning_command(
        cfg,
        msg,
        "set High",
        ambient_context=None,
        topic_store=topic_store,
        chat_prefs=None,
        resolved_scope="main",
        scope_chat_ids=frozenset({123}),
    )

    override = await topic_store.get_engine_override(123, 77, CODEX_ENGINE)
    assert override is not None
    assert override.model == "gpt-4.1-mini"
    assert override.reasoning == "high"
    assert (
        "topic reasoning override set to high for codex."
        in transport.send_calls[-1]["message"].text
    )

    msg_clear = replace(
        msg,
        message_id=11,
        text="/reasoning clear",
    )
    await _handle_reasoning_command(
        cfg,
        msg_clear,
        "clear",
        ambient_context=None,
        topic_store=topic_store,
        chat_prefs=None,
        resolved_scope="main",
        scope_chat_ids=frozenset({123}),
    )

    override = await topic_store.get_engine_override(123, 77, CODEX_ENGINE)
    assert override is not None
    assert override.model == "gpt-4.1-mini"
    assert override.reasoning is None
    assert (
        "topic reasoning override cleared (using chat default)."
        in transport.send_calls[-1]["message"].text
    )


@pytest.mark.anyio
async def test_reasoning_command_show_reports_overrides(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    cfg = replace(cfg, topics=TelegramTopicsSettings(enabled=True, scope="main"))
    chat_prefs = ChatPrefsStore(tmp_path / "telegram_chat_prefs_state.json")
    topic_store = TopicStateStore(tmp_path / "telegram_topics_state.json")
    await chat_prefs.set_engine_override(
        123,
        CODEX_ENGINE,
        EngineOverrides(model=None, reasoning="low"),
    )
    await topic_store.set_engine_override(
        123,
        88,
        CODEX_ENGINE,
        EngineOverrides(model=None, reasoning="high"),
    )
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=10,
        text="/reasoning",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=123,
        thread_id=88,
    )

    await _handle_reasoning_command(
        cfg,
        msg,
        "",
        ambient_context=None,
        topic_store=topic_store,
        chat_prefs=chat_prefs,
        resolved_scope="main",
        scope_chat_ids=frozenset({123}),
    )

    text = transport.send_calls[-1]["message"].text
    assert "engine: codex (global default)" in text
    assert "reasoning: high (topic override)" in text
    assert "defaults: topic: high, chat: low" in text
    assert "available levels: minimal, low, medium, high, xhigh" in text


@pytest.mark.anyio
async def test_send_with_resume_waits_for_token() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    sent: list[
        tuple[
            int,
            int,
            str,
            ResumeToken,
            RunContext | None,
            int | None,
            tuple[int, int | None] | None,
            MessageRef | None,
        ]
    ] = []

    async def enqueue(
        chat_id: int,
        user_msg_id: int,
        text: str,
        resume: ResumeToken,
        context: RunContext | None,
        thread_id: int | None,
        session_key: tuple[int, int | None] | None,
        progress_ref: MessageRef | None,
    ) -> None:
        sent.append(
            (
                chat_id,
                user_msg_id,
                text,
                resume,
                context,
                thread_id,
                session_key,
                progress_ref,
            )
        )

    running_task = RunningTask()

    async def trigger_resume() -> None:
        await anyio.sleep(0)
        running_task.resume = ResumeToken(engine=CODEX_ENGINE, value="abc123")
        running_task.resume_ready.set()

    async with anyio.create_task_group() as tg:
        tg.start_soon(trigger_resume)
        await send_with_resume(
            cfg,
            enqueue,
            running_task,
            123,
            10,
            None,
            None,
            "hello",
        )

    assert len(sent) == 1
    assert sent[0][:7] == (
        123,
        10,
        "hello",
        ResumeToken(engine=CODEX_ENGINE, value="abc123"),
        None,
        None,
        None,
    )
    assert sent[0][7] == transport.send_calls[0]["ref"]
    assert transport.send_calls
    assert "queued" in transport.send_calls[0]["message"].text.lower()


@pytest.mark.anyio
async def test_send_with_resume_reports_when_missing() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    sent: list[
        tuple[
            int,
            int,
            str,
            ResumeToken,
            RunContext | None,
            int | None,
            tuple[int, int | None] | None,
            MessageRef | None,
        ]
    ] = []

    async def enqueue(
        chat_id: int,
        user_msg_id: int,
        text: str,
        resume: ResumeToken,
        context: RunContext | None,
        thread_id: int | None,
        session_key: tuple[int, int | None] | None,
        progress_ref: MessageRef | None,
    ) -> None:
        sent.append(
            (
                chat_id,
                user_msg_id,
                text,
                resume,
                context,
                thread_id,
                session_key,
                progress_ref,
            )
        )

    running_task = RunningTask()
    running_task.done.set()

    await send_with_resume(
        cfg,
        enqueue,
        running_task,
        123,
        10,
        None,
        None,
        "hello",
    )

    assert sent == []
    assert transport.send_calls
    assert "resume token" in transport.send_calls[-1]["message"].text.lower()


@pytest.mark.anyio
async def test_run_engine_hides_resume_line_in_topics() -> None:
    transport = _CaptureTransport()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value="resume-123",
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )

    await _run_engine(
        exec_cfg=exec_cfg,
        runtime=runtime,
        running_tasks={},
        chat_id=123,
        user_msg_id=1,
        text="hello",
        resume_token=None,
        context=None,
        reply_ref=None,
        on_thread_known=None,
        engine_override=None,
        thread_id=77,
        show_resume_line=False,
    )

    assert transport.last_message is not None
    assert "resume-123" not in transport.last_message.text


@pytest.mark.anyio
async def test_run_main_loop_routes_reply_to_running_resume() -> None:
    progress_ready = anyio.Event()
    stop_polling = anyio.Event()
    reply_ready = anyio.Event()
    hold = anyio.Event()

    transport = FakeTransport(progress_ready=progress_ready)
    bot = FakeBot()
    resume_value = "abc123"
    runner = ScriptRunner(
        [Wait(hold), Sleep(0.05), Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="first",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )
        await progress_ready.wait()
        assert transport.progress_ref is not None
        assert isinstance(transport.progress_ref.message_id, int)
        reply_id = transport.progress_ref.message_id
        reply_ready.set()
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="followup",
            reply_to_message_id=reply_id,
            reply_to_text=None,
            sender_id=123,
        )
        await stop_polling.wait()

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_main_loop, cfg, poller)
        try:
            with anyio.fail_after(2):
                await reply_ready.wait()
            await anyio.sleep(0)
            hold.set()
            with anyio.fail_after(2):
                while len(runner.calls) < 2:
                    await anyio.sleep(0)
            assert runner.calls[1][1] == ResumeToken(
                engine=CODEX_ENGINE, value=resume_value
            )
        finally:
            hold.set()
            stop_polling.set()
            tg.cancel_scope.cancel()


@pytest.mark.anyio
async def test_run_main_loop_persists_topic_sessions_in_project_scope(
    tmp_path: Path,
) -> None:
    project_chat_id = -100
    resume_value = "resume-123"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    projects = ProjectsConfig(
        projects={
            "yee88": ProjectConfig(
                alias="yee88",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
                chat_id=project_chat_id,
            )
        },
        default_project=None,
        chat_map={project_chat_id: "yee88"},
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=projects,
        config_path=tmp_path / "yee88.toml",
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        topics=TelegramTopicsSettings(
            enabled=True,
            scope="projects",
        ),
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=project_chat_id,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            thread_id=77,
        )

    with anyio.fail_after(2):
        await run_main_loop(cfg, poller)

    state_path = resolve_state_path(runtime.config_path or tmp_path / "yee88.toml")
    store = TopicStateStore(state_path)
    stored = await store.get_session_resume(project_chat_id, 77, CODEX_ENGINE)
    assert stored == ResumeToken(engine=CODEX_ENGINE, value=resume_value)


@pytest.mark.anyio
async def test_run_main_loop_auto_resumes_topic_default_engine(
    tmp_path: Path,
) -> None:
    state_path = tmp_path / "yee88.toml"
    topic_path = resolve_state_path(state_path)
    store = TopicStateStore(topic_path)
    await store.set_session_resume(
        123, 77, ResumeToken(engine=CODEX_ENGINE, value="resume-codex")
    )
    await store.set_session_resume(
        123, 77, ResumeToken(engine="claude", value="resume-claude")
    )
    await store.set_default_engine(123, 77, "claude")

    transport = FakeTransport()
    bot = FakeBot()
    codex_runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    claude_runner = ScriptRunner([Return(answer="ok")], engine="claude")
    router = AutoRouter(
        entries=[
            RunnerEntry(engine=codex_runner.engine, runner=codex_runner),
            RunnerEntry(engine=claude_runner.engine, runner=claude_runner),
        ],
        default_engine=codex_runner.engine,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
                chat_id=123,
            )
        },
        default_project=None,
        chat_map={123: "proj"},
    )
    runtime = TransportRuntime(
        router=router,
        projects=projects,
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=ExecBridgeConfig(
            transport=transport,
            presenter=MarkdownPresenter(),
            final_notify=True,
        ),
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        topics=TelegramTopicsSettings(
            enabled=True,
            scope="main",
        ),
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            thread_id=77,
        )

    await run_main_loop(cfg, poller)

    assert codex_runner.calls == []
    assert len(claude_runner.calls) == 1
    assert claude_runner.calls[0][1] == ResumeToken(
        engine="claude", value="resume-claude"
    )


@pytest.mark.anyio
async def test_run_main_loop_auto_resumes_chat_sessions(tmp_path: Path) -> None:
    resume_value = "resume-123"
    state_path = tmp_path / "yee88.toml"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project="proj",
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=projects,
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg, poller)

    store = ChatSessionStore(resolve_sessions_path(state_path))
    stored = await store.get_session_resume(123, None, CODEX_ENGINE)
    assert stored == ResumeToken(engine=CODEX_ENGINE, value=resume_value)

    runner2 = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    runtime2 = TransportRuntime(
        router=_make_router(runner2),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg2 = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime2,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
    )

    async def poller2(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="followup",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg2, poller2)

    assert runner2.calls[0][1] == ResumeToken(engine=CODEX_ENGINE, value=resume_value)


@pytest.mark.anyio
async def test_run_main_loop_prompt_upload_uses_caption_directives(
    tmp_path: Path,
) -> None:
    payload = b"hello"
    proj_dir = tmp_path / "proj"
    other_dir = tmp_path / "other"
    proj_dir.mkdir()
    other_dir.mkdir()

    class _UploadBot(FakeBot):
        async def get_file(self, file_id: str) -> File | None:
            _ = file_id
            return File(file_path="files/hello.txt")

        async def download_file(self, file_path: str) -> bytes | None:
            _ = file_path
            return payload

    transport = FakeTransport()
    bot = _UploadBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=proj_dir,
                worktrees_dir=Path(".worktrees"),
            ),
            "other": ProjectConfig(
                alias="other",
                path=other_dir,
                worktrees_dir=Path(".worktrees"),
            ),
        },
        default_project="proj",
    )
    runtime = TransportRuntime(router=_make_router(runner), projects=projects)
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        files=TelegramFilesSettings(
            enabled=True,
            auto_put=True,
            auto_put_mode="prompt",
        ),
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/other do thing",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
            document=TelegramDocument(
                file_id="doc-1",
                file_name="hello.txt",
                mime_type="text/plain",
                file_size=len(payload),
                raw={"file_id": "doc-1"},
            ),
        )

    await run_main_loop(cfg, poller)

    saved_path = other_dir / "incoming" / "hello.txt"
    assert saved_path.read_bytes() == payload
    assert runner.calls
    prompt_text, _ = runner.calls[0]
    assert prompt_text.startswith("do thing")
    assert "/other" not in prompt_text
    assert "[uploaded file: incoming/hello.txt]" in prompt_text


@pytest.mark.anyio
async def test_run_main_loop_voice_transcript_preserves_directive(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    codex_runner = ScriptRunner([Return(answer="codex")], engine=CODEX_ENGINE)
    claude_runner = ScriptRunner([Return(answer="claude")], engine="claude")
    router = AutoRouter(
        entries=[
            RunnerEntry(engine=claude_runner.engine, runner=claude_runner),
            RunnerEntry(engine=codex_runner.engine, runner=codex_runner),
        ],
        default_engine=claude_runner.engine,
    )
    runtime = TransportRuntime(router=router, projects=_empty_projects())
    transport = FakeTransport()
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=FakeBot(),
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        voice_transcription=True,
    )

    async def _fake_transcribe(
        *,
        bot: BotClient,
        msg: TelegramIncomingMessage,
        enabled: bool,
        model: str,
        max_bytes: int | None = None,
        reply,
        base_url: str | None = None,
        api_key: str | None = None,
    ) -> str:
        _ = bot, msg, enabled, model, max_bytes, reply, base_url, api_key
        return "/codex do thing"

    monkeypatch.setattr(telegram_loop, "transcribe_voice", _fake_transcribe)
    monkeypatch.setattr(telegram_loop, "list_command_ids", lambda **_: [])

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            voice=TelegramVoice(
                file_id="voice-1",
                mime_type=None,
                file_size=None,
                duration=None,
                raw={"file_id": "voice-1"},
            ),
        )

    await run_main_loop(cfg, poller)

    assert not claude_runner.calls
    assert len(codex_runner.calls) == 1
    assert codex_runner.calls[0][0].startswith("(voice transcribed) do thing")


@pytest.mark.anyio
async def test_run_main_loop_debounces_forwarded_messages_preserves_directives() -> (
    None
):
    codex_runner = ScriptRunner([Return(answer="codex")], engine=CODEX_ENGINE)
    claude_runner = ScriptRunner([Return(answer="claude")], engine="claude")
    router = AutoRouter(
        entries=[
            RunnerEntry(engine=claude_runner.engine, runner=claude_runner),
            RunnerEntry(engine=codex_runner.engine, runner=codex_runner),
        ],
        default_engine=claude_runner.engine,
    )
    runtime = TransportRuntime(router=router, projects=_empty_projects())
    transport = FakeTransport()
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=FakeBot(),
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=DEBOUNCE_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/codex summarize these",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )
        await anyio.sleep(_cfg.forward_coalesce_s / 2)
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="a",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            raw={"forward_origin": {"type": "user"}},
        )
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=3,
            text="b",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            raw={"forward_origin": {"type": "user"}},
        )
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=4,
            text="c",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            raw={"forward_origin": {"type": "user"}},
        )

    await run_main_loop(cfg, poller)

    assert not claude_runner.calls
    assert len(codex_runner.calls) == 1
    prompt_text, _ = codex_runner.calls[0]
    assert prompt_text == "summarize these\n\na\n\nb\n\nc"


@pytest.mark.anyio
async def test_run_main_loop_ignores_forwarded_without_prompt() -> None:
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    runtime = TransportRuntime(router=_make_router(runner), projects=_empty_projects())
    transport = FakeTransport()
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=FakeBot(),
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="a",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            raw={"forward_origin": {"type": "user"}},
        )
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="b",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            raw={"forward_origin": {"type": "user"}},
        )

    await run_main_loop(cfg, poller)

    assert runner.calls == []


@pytest.mark.anyio
async def test_run_main_loop_forwarded_document_still_uploads(
    tmp_path: Path,
) -> None:
    payload = b"hello"

    class _UploadBot(FakeBot):
        async def get_file(self, file_id: str) -> File | None:
            _ = file_id
            return File(file_path="files/hello.txt")

        async def download_file(self, file_path: str) -> bytes | None:
            _ = file_path
            return payload

    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project="proj",
    )
    runtime = TransportRuntime(router=_make_router(runner), projects=projects)
    transport = FakeTransport()
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=_UploadBot(),
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        files=TelegramFilesSettings(
            enabled=True,
            auto_put=True,
            auto_put_mode="prompt",
        ),
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="do thing",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
            document=TelegramDocument(
                file_id="doc-1",
                file_name="hello.txt",
                mime_type="text/plain",
                file_size=len(payload),
                raw={"file_id": "doc-1"},
            ),
            raw={"forward_origin": {"type": "user"}},
        )

    await run_main_loop(cfg, poller)

    saved_path = tmp_path / "incoming" / "hello.txt"
    assert saved_path.read_bytes() == payload
    assert runner.calls
    prompt_text, _ = runner.calls[0]
    assert prompt_text.startswith("do thing")
    assert "[uploaded file: incoming/hello.txt]" in prompt_text


@pytest.mark.anyio
async def test_run_main_loop_prompt_upload_auto_resumes_chat_sessions(
    tmp_path: Path,
) -> None:
    payload = b"hello"
    resume_value = "resume-123"
    state_path = tmp_path / "yee88.toml"
    project_dir = tmp_path / "proj"
    project_dir.mkdir()

    class _UploadBot(FakeBot):
        async def get_file(self, file_id: str) -> File | None:
            _ = file_id
            return File(file_path="files/hello.txt")

        async def download_file(self, file_path: str) -> bytes | None:
            _ = file_path
            return payload

    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=project_dir,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project="proj",
    )
    bot = _UploadBot()

    transport = FakeTransport()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=projects,
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        files=TelegramFilesSettings(
            enabled=True,
            auto_put=True,
            auto_put_mode="prompt",
        ),
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
            document=TelegramDocument(
                file_id="doc-1",
                file_name="hello.txt",
                mime_type="text/plain",
                file_size=len(payload),
                raw={"file_id": "doc-1"},
            ),
        )

    await run_main_loop(cfg, poller)

    store = ChatSessionStore(resolve_sessions_path(state_path))
    stored = await store.get_session_resume(123, None, CODEX_ENGINE)
    assert stored == ResumeToken(engine=CODEX_ENGINE, value=resume_value)

    transport2 = FakeTransport()
    runner2 = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg2 = ExecBridgeConfig(
        transport=transport2,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime2 = TransportRuntime(
        router=_make_router(runner2),
        projects=projects,
        config_path=state_path,
    )
    cfg2 = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime2,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg2,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        files=TelegramFilesSettings(
            enabled=True,
            auto_put=True,
            auto_put_mode="prompt",
        ),
    )

    async def poller2(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="followup",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
            document=TelegramDocument(
                file_id="doc-2",
                file_name="hello2.txt",
                mime_type="text/plain",
                file_size=len(payload),
                raw={"file_id": "doc-2"},
            ),
        )

    await run_main_loop(cfg2, poller2)

    assert runner2.calls[0][1] == ResumeToken(
        engine=CODEX_ENGINE,
        value=resume_value,
    )


@pytest.mark.anyio
async def test_run_main_loop_command_updates_chat_session_resume(
    tmp_path: Path,
    monkeypatch,
) -> None:
    class _Command:
        id = "run_cmd"
        description = "run command"

        async def handle(self, ctx):
            await ctx.executor.run_one(commands.RunRequest(prompt="hello"))
            return commands.CommandResult(text="done")

    entrypoints = [
        FakeEntryPoint(
            "run_cmd",
            "yee88.commands.run_cmd:BACKEND",
            plugins.COMMAND_GROUP,
            loader=_Command,
        )
    ]
    install_entrypoints(monkeypatch, entrypoints)

    resume_value = "resume-123"
    state_path = tmp_path / "yee88.toml"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        show_resume_line=False,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/run_cmd",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg, poller)

    store = ChatSessionStore(resolve_sessions_path(state_path))
    stored = await store.get_session_resume(123, None, CODEX_ENGINE)
    assert stored == ResumeToken(engine=CODEX_ENGINE, value=resume_value)

    transport2 = FakeTransport()
    runner2 = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg2 = ExecBridgeConfig(
        transport=transport2,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime2 = TransportRuntime(
        router=_make_router(runner2),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg2 = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime2,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg2,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        show_resume_line=False,
    )

    async def poller2(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="followup",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg2, poller2)

    assert runner2.calls[0][1] == ResumeToken(
        engine=CODEX_ENGINE,
        value=resume_value,
    )


@pytest.mark.anyio
async def test_run_main_loop_hides_resume_line_when_disabled(
    tmp_path: Path,
) -> None:
    resume_value = "resume-123"
    state_path = tmp_path / "yee88.toml"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project="proj",
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=projects,
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        show_resume_line=False,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg, poller)

    assert transport.send_calls
    final_text = transport.send_calls[-1]["message"].text
    assert resume_value not in final_text


@pytest.mark.anyio
async def test_run_main_loop_hides_resume_line_without_context(
    tmp_path: Path,
) -> None:
    resume_value = "resume-ctxless"
    state_path = tmp_path / "yee88.toml"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        show_resume_line=False,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg, poller)

    assert transport.send_calls
    final_text = transport.send_calls[-1]["message"].text
    assert resume_value not in final_text


@pytest.mark.anyio
async def test_run_main_loop_applies_chat_bound_context(
    tmp_path: Path,
) -> None:
    state_path = tmp_path / "yee88.toml"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    projects = ProjectsConfig(
        projects={
            "alpha": ProjectConfig(
                alias="Alpha",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            ),
            "beta": ProjectConfig(
                alias="Beta",
                path=tmp_path / "beta",
                worktrees_dir=Path(".worktrees"),
            ),
        },
        default_project="alpha",
    )
    (tmp_path / "beta").mkdir()
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=projects,
        config_path=state_path,
    )
    prefs = ChatPrefsStore(resolve_prefs_path(state_path))
    await prefs.set_context(123, RunContext(project="beta"))
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
        show_resume_line=False,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg, poller)

    assert transport.send_calls
    final_text = transport.send_calls[-1]["message"].text
    assert "`ctx: Beta`" in final_text


@pytest.mark.anyio
async def test_run_main_loop_chat_sessions_isolate_group_senders(
    tmp_path: Path,
) -> None:
    resume_value = "resume-group"
    state_path = tmp_path / "yee88.toml"

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner(
        [Return(answer="ok")],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=-100,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=111,
            chat_type="supergroup",
        )

    await run_main_loop(cfg, poller)

    runner2 = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    runtime2 = TransportRuntime(
        router=_make_router(runner2),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg2 = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime2,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
    )

    async def poller2(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=-100,
            message_id=2,
            text="followup",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=222,
            chat_type="supergroup",
        )

    await run_main_loop(cfg2, poller2)

    assert runner2.calls[0][1] is None


@pytest.mark.anyio
async def test_run_main_loop_new_clears_chat_sessions(tmp_path: Path) -> None:
    state_path = tmp_path / "yee88.toml"
    store = ChatSessionStore(resolve_sessions_path(state_path))
    await store.set_session_resume(
        123, None, ResumeToken(engine=CODEX_ENGINE, value="resume-1")
    )

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        session_mode="chat",
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/new",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            chat_type="private",
        )

    await run_main_loop(cfg, poller)

    store2 = ChatSessionStore(resolve_sessions_path(state_path))
    assert await store2.get_session_resume(123, None, CODEX_ENGINE) is None


@pytest.mark.anyio
async def test_run_main_loop_new_clears_topic_sessions(tmp_path: Path) -> None:
    state_path = tmp_path / "yee88.toml"
    store = TopicStateStore(resolve_state_path(state_path))
    await store.set_session_resume(
        123, 77, ResumeToken(engine=CODEX_ENGINE, value="resume-1")
    )

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
        config_path=state_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        topics=TelegramTopicsSettings(enabled=True, scope="main"),
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/new",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            thread_id=77,
            chat_type="supergroup",
        )

    with anyio.fail_after(2):
        await run_main_loop(cfg, poller)

    store2 = TopicStateStore(resolve_state_path(state_path))
    assert await store2.get_session_resume(123, 77, CODEX_ENGINE) is None


@pytest.mark.anyio
async def test_run_main_loop_replies_in_same_thread() -> None:
    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            thread_id=77,
        )

    await run_main_loop(cfg, poller)

    reply_calls = [
        call
        for call in transport.send_calls
        if call["options"] is not None and call["options"].reply_to is not None
    ]
    assert reply_calls
    assert all(call["options"].thread_id == 77 for call in reply_calls)


@pytest.mark.anyio
async def test_run_main_loop_batches_media_group_upload(
    tmp_path: Path,
) -> None:
    payloads = {
        "photos/file_1.jpg": b"one",
        "photos/file_2.jpg": b"two",
    }
    file_map = {
        "doc-1": "photos/file_1.jpg",
        "doc-2": "photos/file_2.jpg",
    }

    class _MediaBot(FakeBot):
        async def get_file(self, file_id: str) -> File | None:
            file_path = file_map.get(file_id)
            if file_path is None:
                return None
            return File(file_path=file_path)

        async def download_file(self, file_path: str) -> bytes | None:
            return payloads.get(file_path)

    transport = FakeTransport()
    bot = _MediaBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=tmp_path,
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project=None,
    )
    runtime = TransportRuntime(router=_make_router(runner), projects=projects)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=BATCH_MEDIA_GROUP_DEBOUNCE_S,
        files=TelegramFilesSettings(enabled=True, auto_put=True),
    )
    msg1 = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=1,
        text="/file put /proj incoming/test1",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=321,
        chat_type="private",
        media_group_id="grp-1",
        document=TelegramDocument(
            file_id="doc-1",
            file_name=None,
            mime_type="image/jpeg",
            file_size=len(payloads["photos/file_1.jpg"]),
            raw={"file_id": "doc-1"},
        ),
    )
    msg2 = TelegramIncomingMessage(
        transport="telegram",
        chat_id=123,
        message_id=2,
        text="",
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=321,
        chat_type="private",
        media_group_id="grp-1",
        document=TelegramDocument(
            file_id="doc-2",
            file_name=None,
            mime_type="image/jpeg",
            file_size=len(payloads["photos/file_2.jpg"]),
            raw={"file_id": "doc-2"},
        ),
    )

    stop_polling = anyio.Event()

    async def poller(_cfg: TelegramBridgeConfig):
        yield msg1
        yield msg2
        await stop_polling.wait()

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_main_loop, cfg, poller)
        try:
            with anyio.fail_after(3):
                while len(transport.send_calls) < 1:
                    await anyio.sleep(0.05)
            assert len(transport.send_calls) == 1
            text = transport.send_calls[0]["message"].text
            assert "saved file_1.jpg, file_2.jpg" in text
            assert "to incoming/test1/" in text
            target_dir = tmp_path / "incoming" / "test1"
            assert (target_dir / "file_1.jpg").read_bytes() == payloads[
                "photos/file_1.jpg"
            ]
            assert (target_dir / "file_2.jpg").read_bytes() == payloads[
                "photos/file_2.jpg"
            ]
        finally:
            stop_polling.set()
            tg.cancel_scope.cancel()


@pytest.mark.anyio
async def test_run_main_loop_handles_command_plugins(monkeypatch) -> None:
    class _Command:
        id = "echo_cmd"
        description = "echo"

        async def handle(self, ctx):
            return commands.CommandResult(text=f"echo:{ctx.args_text}")

    entrypoints = [
        FakeEntryPoint(
            "echo_cmd",
            "yee88.commands.echo:BACKEND",
            plugins.COMMAND_GROUP,
            loader=_Command,
        )
    ]
    install_entrypoints(monkeypatch, entrypoints)

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/echo_cmd hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert runner.calls == []
    assert transport.send_calls
    assert transport.send_calls[-1]["message"].text == "echo:hello"


@pytest.mark.anyio
async def test_run_main_loop_command_uses_project_default_engine(
    monkeypatch,
) -> None:
    class _Command:
        id = "use_project"
        description = "use project default"

        async def handle(self, ctx):
            result = await ctx.executor.run_one(
                commands.RunRequest(
                    prompt="hello",
                    context=RunContext(project="proj"),
                ),
                mode="capture",
            )
            return commands.CommandResult(text=f"ran:{result.engine}")

    entrypoints = [
        FakeEntryPoint(
            "use_project",
            "yee88.commands.use_project:BACKEND",
            plugins.COMMAND_GROUP,
            loader=_Command,
        )
    ]
    install_entrypoints(monkeypatch, entrypoints)

    transport = FakeTransport()
    bot = FakeBot()
    codex_runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    pi_runner = ScriptRunner([Return(answer="ok")], engine="pi")
    router = AutoRouter(
        entries=[
            RunnerEntry(engine=codex_runner.engine, runner=codex_runner),
            RunnerEntry(engine=pi_runner.engine, runner=pi_runner),
        ],
        default_engine=codex_runner.engine,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
                default_engine=pi_runner.engine,
            )
        },
        default_project=None,
    )
    runtime = TransportRuntime(
        router=router,
        projects=projects,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/use_project",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert codex_runner.calls == []
    assert len(pi_runner.calls) == 1
    assert transport.send_calls[-1]["message"].text == "ran:pi"


@pytest.mark.anyio
async def test_run_main_loop_command_defaults_to_chat_project(
    monkeypatch,
) -> None:
    class _Command:
        id = "auto_ctx"
        description = "auto context"

        async def handle(self, ctx):
            result = await ctx.executor.run_one(
                commands.RunRequest(prompt="hello"),
                mode="capture",
            )
            return commands.CommandResult(text=f"ran:{result.engine}")

    entrypoints = [
        FakeEntryPoint(
            "auto_ctx",
            "yee88.commands.auto_ctx:BACKEND",
            plugins.COMMAND_GROUP,
            loader=_Command,
        )
    ]
    install_entrypoints(monkeypatch, entrypoints)

    transport = FakeTransport()
    bot = FakeBot()
    codex_runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    pi_runner = ScriptRunner([Return(answer="ok")], engine="pi")
    router = AutoRouter(
        entries=[
            RunnerEntry(engine=codex_runner.engine, runner=codex_runner),
            RunnerEntry(engine=pi_runner.engine, runner=pi_runner),
        ],
        default_engine=codex_runner.engine,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
                default_engine=pi_runner.engine,
                chat_id=-42,
            )
        },
        default_project=None,
        chat_map={-42: "proj"},
    )
    runtime = TransportRuntime(
        router=router,
        projects=projects,
    )
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=-42,
            message_id=1,
            text="/auto_ctx",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert codex_runner.calls == []
    assert len(pi_runner.calls) == 1
    assert transport.send_calls[-1]["message"].text == "ran:pi"


@pytest.mark.anyio
async def test_run_main_loop_refreshes_command_ids(monkeypatch) -> None:
    class _Command:
        id = "late_cmd"
        description = "late command"

        async def handle(self, ctx):
            return commands.CommandResult(text="late")

    entrypoints = [
        FakeEntryPoint(
            "late_cmd",
            "yee88.commands.late:BACKEND",
            plugins.COMMAND_GROUP,
            loader=_Command,
        )
    ]
    install_entrypoints(monkeypatch, entrypoints)

    calls = {"count": 0}

    def _list_command_ids(*, allowlist=None):
        _ = allowlist
        calls["count"] += 1
        if calls["count"] == 1:
            return []
        return ["late_cmd"]

    monkeypatch.setattr(telegram_loop, "list_command_ids", _list_command_ids)

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="/late_cmd hello",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
        )

    await run_main_loop(cfg, poller)

    assert calls["count"] >= 2
    assert transport.send_calls[-1]["message"].text == "late"


@pytest.mark.anyio
async def test_run_main_loop_mentions_only_skips_voice_and_files(
    monkeypatch, tmp_path
) -> None:
    calls = {"voice": 0, "file": 0}

    async def fake_transcribe_voice(**kwargs):
        _ = kwargs
        calls["voice"] += 1
        return "hello"

    async def fake_handle_file_put_default(*args, **kwargs):
        _ = args, kwargs
        calls["file"] += 1
        return None

    monkeypatch.setattr(telegram_loop, "transcribe_voice", fake_transcribe_voice)
    monkeypatch.setattr(
        telegram_loop, "_handle_file_put_default", fake_handle_file_put_default
    )

    transport = FakeTransport()
    bot = FakeBot()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    config_path = tmp_path / "yee88.toml"
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
        config_path=config_path,
    )
    cfg = TelegramBridgeConfig(
        bot=bot,
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=FAST_FORWARD_COALESCE_S,
        media_group_debounce_s=FAST_MEDIA_GROUP_DEBOUNCE_S,
        voice_transcription=True,
        files=TelegramFilesSettings(enabled=True, auto_put=True),
    )

    prefs = ChatPrefsStore(resolve_prefs_path(config_path))
    await prefs.set_trigger_mode(123, "mentions")

    voice = TelegramVoice(
        file_id="voice-id",
        mime_type="audio/ogg",
        file_size=5,
        duration=1,
        raw={},
    )
    document = TelegramDocument(
        file_id="doc-id",
        file_name="doc.txt",
        mime_type="text/plain",
        file_size=5,
        raw={},
    )

    async def poller(_cfg: TelegramBridgeConfig):
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=1,
            text="",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            voice=voice,
            raw={},
        )
        yield TelegramIncomingMessage(
            transport="telegram",
            chat_id=123,
            message_id=2,
            text="",
            reply_to_message_id=None,
            reply_to_text=None,
            sender_id=123,
            document=document,
            raw={},
        )

    await run_main_loop(cfg, poller)

    assert calls["voice"] == 0
    assert calls["file"] == 0
    assert runner.calls == []



from pathlib import Path

import pytest

import yee88.runtime_loader as runtime_loader
from yee88.config import ConfigError
from yee88.settings import TakopiSettings


def test_build_runtime_spec_minimal(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(runtime_loader.shutil, "which", lambda _cmd: "/bin/echo")
    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "watch_config": True,
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        }
    )
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n\n[transports.telegram]\n'
        'bot_token = "token"\nchat_id = 123\n',
        encoding="utf-8",
    )

    spec = runtime_loader.build_runtime_spec(
        settings=settings,
        config_path=config_path,
    )

    assert spec.router.default_engine == settings.default_engine
    runtime = spec.to_runtime(config_path=config_path)
    assert runtime.default_engine == settings.default_engine
    assert runtime.watch_config is True


def test_resolve_default_engine_unknown(tmp_path: Path) -> None:
    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        }
    )
    with pytest.raises(ConfigError, match="Unknown default engine"):
        runtime_loader.resolve_default_engine(
            override="unknown",
            settings=settings,
            config_path=tmp_path / "yee88.toml",
            engine_ids=["codex"],
        )



from pathlib import Path

import anyio
import pytest

import yee88.config_watch as config_watch
from yee88.config_watch import ConfigReload, config_status, watch_config
from yee88.config import ProjectsConfig
from yee88.router import AutoRouter, RunnerEntry
from yee88.runtime_loader import RuntimeSpec
from yee88.runners.mock import Return, ScriptRunner
from yee88.settings import TakopiSettings
from yee88.transport_runtime import TransportRuntime


def test_config_status_variants(tmp_path: Path) -> None:
    missing = tmp_path / "missing.toml"
    status, signature = config_status(missing)
    assert status == "missing"
    assert signature is None

    directory = tmp_path / "config.d"
    directory.mkdir()
    status, signature = config_status(directory)
    assert status == "invalid"
    assert signature is None

    config_file = tmp_path / "yee88.toml"
    config_file.write_text(
        'transport = "telegram"\n\n[transports.telegram]\n'
        'bot_token = "token"\nchat_id = 123\n',
        encoding="utf-8",
    )
    status, signature = config_status(config_file)
    assert status == "ok"
    assert signature is not None


@pytest.mark.anyio
async def test_watch_config_applies_runtime(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text('default_engine = "codex"\n', encoding="utf-8")
    resolved_path = config_path.resolve()

    codex_runner = ScriptRunner([Return(answer="ok")], engine="codex")
    router = AutoRouter(
        entries=[RunnerEntry(engine=codex_runner.engine, runner=codex_runner)],
        default_engine=codex_runner.engine,
    )
    runtime = TransportRuntime(
        router=router,
        projects=ProjectsConfig(projects={}, default_project=None),
        config_path=resolved_path,
    )

    pi_runner = ScriptRunner([Return(answer="ok")], engine="pi")
    new_router = AutoRouter(
        entries=[RunnerEntry(engine=pi_runner.engine, runner=pi_runner)],
        default_engine=pi_runner.engine,
    )
    new_spec = RuntimeSpec(
        router=new_router,
        projects=ProjectsConfig(projects={}, default_project=None),
        allowlist=None,
        plugin_configs=None,
    )
    reload = ConfigReload(
        settings=TakopiSettings.model_validate(
            {
                "transport": "telegram",
                "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
            }
        ),
        runtime_spec=new_spec,
        config_path=resolved_path,
    )

    ready = anyio.Event()
    watching = anyio.Event()

    async def fake_awatch(_path: Path):
        watching.set()
        await ready.wait()
        yield {(None, str(resolved_path))}

    monkeypatch.setattr(config_watch, "awatch", fake_awatch)
    monkeypatch.setattr(
        config_watch, "_reload_config", lambda *_args, **_kwargs: reload
    )

    reloaded = anyio.Event()

    async def on_reload(_payload: ConfigReload) -> None:
        reloaded.set()

    async with anyio.create_task_group() as tg:

        async def run_watch() -> None:
            await watch_config(
                config_path=resolved_path,
                runtime=runtime,
                on_reload=on_reload,
            )

        tg.start_soon(run_watch)
        with anyio.fail_after(2):
            await watching.wait()
        config_path.write_text('default_engine = "pi"\n', encoding="utf-8")
        ready.set()
        with anyio.fail_after(2):
            await reloaded.wait()
        tg.cancel_scope.cancel()

    assert runtime.default_engine == "pi"



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Protocol, runtime_checkable
from collections.abc import Iterable

from .backends import EngineBackend, SetupIssue
from .plugins import TRANSPORT_GROUP, list_ids, load_plugin_backend
from .transport_runtime import TransportRuntime


@dataclass(frozen=True, slots=True)
class SetupResult:
    issues: list[SetupIssue]
    config_path: Path

    @property
    def ok(self) -> bool:
        return not self.issues


@runtime_checkable
class TransportBackend(Protocol):
    id: str
    description: str

    def check_setup(
        self,
        engine_backend: EngineBackend,
        *,
        transport_override: str | None = None,
    ) -> SetupResult: ...

    async def interactive_setup(self, *, force: bool) -> bool: ...

    def lock_token(
        self, *, transport_config: object, _config_path: Path
    ) -> str | None: ...

    def build_and_run(
        self,
        *,
        transport_config: object,
        config_path: Path,
        runtime: TransportRuntime,
        final_notify: bool,
        default_engine_override: str | None,
    ) -> None: ...


def _validate_transport_backend(backend: object, ep) -> None:
    if not isinstance(backend, TransportBackend):
        raise TypeError(f"{ep.value} is not a TransportBackend")
    if backend.id != ep.name:
        raise ValueError(
            f"{ep.value} transport id {backend.id!r} does not match entrypoint {ep.name!r}"
        )


def get_transport(
    transport_id: str, *, allowlist: Iterable[str] | None = None
) -> TransportBackend:
    backend = load_plugin_backend(
        TRANSPORT_GROUP,
        transport_id,
        allowlist=allowlist,
        validator=_validate_transport_backend,
        kind_label="transport",
    )
    assert backend is not None
    return backend


def list_transports(*, allowlist: Iterable[str] | None = None) -> list[str]:
    return list_ids(TRANSPORT_GROUP, allowlist=allowlist)



"""Msgspec models for Telegram Bot API payloads (subset used by yee88).

Derived from telegram-api.html in the repository.
"""

from __future__ import annotations

import msgspec

__all__ = [
    "CallbackQuery",
    "CallbackQueryMessage",
    "Chat",
    "ChatMember",
    "Document",
    "File",
    "ForumTopic",
    "Message",
    "MessageReply",
    "PhotoSize",
    "Sticker",
    "Update",
    "User",
    "Video",
    "Voice",
    "decode_update",
    "decode_updates",
]


class User(msgspec.Struct, forbid_unknown_fields=False):
    id: int
    is_bot: bool | None = None
    username: str | None = None
    first_name: str | None = None
    last_name: str | None = None


class Chat(msgspec.Struct, forbid_unknown_fields=False):
    id: int
    type: str
    title: str | None = None
    username: str | None = None
    first_name: str | None = None
    last_name: str | None = None
    is_forum: bool | None = None


class PhotoSize(msgspec.Struct, forbid_unknown_fields=False):
    file_id: str
    width: int
    height: int
    file_size: int | None = None


class Document(msgspec.Struct, forbid_unknown_fields=False):
    file_id: str
    file_name: str | None = None
    mime_type: str | None = None
    file_size: int | None = None


class Video(msgspec.Struct, forbid_unknown_fields=False):
    file_id: str
    file_name: str | None = None
    mime_type: str | None = None
    file_size: int | None = None


class Voice(msgspec.Struct, forbid_unknown_fields=False):
    file_id: str
    duration: int | None = None
    mime_type: str | None = None
    file_size: int | None = None


class Sticker(msgspec.Struct, forbid_unknown_fields=False):
    file_id: str
    file_size: int | None = None


class MessageReply(msgspec.Struct, forbid_unknown_fields=False):
    message_id: int
    text: str | None = None
    from_: User | None = msgspec.field(default=None, name="from")


class Message(msgspec.Struct, forbid_unknown_fields=False):
    message_id: int
    chat: Chat
    message_thread_id: int | None = None
    from_: User | None = msgspec.field(default=None, name="from")
    text: str | None = None
    caption: str | None = None
    reply_to_message: MessageReply | None = None
    forward_from: User | None = None
    forward_from_chat: Chat | None = None
    forward_from_message_id: int | None = None
    forward_sender_name: str | None = None
    forward_signature: str | None = None
    forward_date: int | None = None
    media_group_id: str | None = None
    is_automatic_forward: bool | None = None
    is_topic_message: bool | None = None
    voice: Voice | None = None
    document: Document | None = None
    video: Video | None = None
    photo: list[PhotoSize] | None = None
    sticker: Sticker | None = None


class CallbackQueryMessage(msgspec.Struct, forbid_unknown_fields=False):
    message_id: int
    chat: Chat


class CallbackQuery(msgspec.Struct, forbid_unknown_fields=False):
    id: str
    from_: User = msgspec.field(name="from")
    message: CallbackQueryMessage | None = None
    data: str | None = None


class Update(msgspec.Struct, forbid_unknown_fields=False):
    update_id: int
    message: Message | None = None
    callback_query: CallbackQuery | None = None


class File(msgspec.Struct, forbid_unknown_fields=False):
    file_path: str


class ChatMember(msgspec.Struct, forbid_unknown_fields=False):
    status: str
    can_manage_topics: bool | None = None


class ForumTopic(msgspec.Struct, forbid_unknown_fields=False):
    message_thread_id: int


_UPDATE_DECODER = msgspec.json.Decoder(Update)
_UPDATES_DECODER = msgspec.json.Decoder(list[Update])


def decode_update(payload: str | bytes) -> Update:
    return _UPDATE_DECODER.decode(payload)


def decode_updates(payload: str | bytes) -> list[Update]:
    return _UPDATES_DECODER.decode(payload)



from typing import Any

import anyio

from yee88.config import ProjectsConfig
from yee88.markdown import MarkdownPresenter
from yee88.router import AutoRouter, RunnerEntry
from yee88.runner_bridge import ExecBridgeConfig
from yee88.runners.mock import Return, ScriptRunner
from yee88.telegram.api_models import (
    Chat,
    ChatMember,
    File,
    ForumTopic,
    Message,
    Update,
    User,
)
from yee88.telegram.bridge import TelegramBridgeConfig
from yee88.telegram.client import BotClient
from yee88.transport import MessageRef, RenderedMessage, SendOptions
from yee88.transport_runtime import TransportRuntime

DEFAULT_ENGINE_ID = "codex"


def _empty_projects() -> ProjectsConfig:
    return ProjectsConfig(projects={}, default_project=None)


def _make_router(runner: Any) -> AutoRouter:
    return AutoRouter(
        entries=[RunnerEntry(engine=runner.engine, runner=runner)],
        default_engine=runner.engine,
    )


class FakeTransport:
    def __init__(self, progress_ready: anyio.Event | None = None) -> None:
        self._next_id = 1
        self.send_calls: list[dict] = []
        self.edit_calls: list[dict] = []
        self.delete_calls: list[MessageRef] = []
        self.progress_ready = progress_ready
        self.progress_ref: MessageRef | None = None

    async def send(
        self,
        *,
        channel_id: int | str,
        message: RenderedMessage,
        options: SendOptions | None = None,
    ) -> MessageRef:
        ref = MessageRef(channel_id=channel_id, message_id=self._next_id)
        self._next_id += 1
        self.send_calls.append(
            {
                "ref": ref,
                "channel_id": channel_id,
                "message": message,
                "options": options,
            }
        )
        if (
            self.progress_ref is None
            and options is not None
            and options.reply_to is not None
            and options.notify is False
        ):
            self.progress_ref = ref
            if self.progress_ready is not None:
                self.progress_ready.set()
        return ref

    async def edit(
        self, *, ref: MessageRef, message: RenderedMessage, wait: bool = True
    ) -> MessageRef:
        self.edit_calls.append({"ref": ref, "message": message, "wait": wait})
        return ref

    async def delete(self, *, ref: MessageRef) -> bool:
        self.delete_calls.append(ref)
        return True

    async def close(self) -> None:
        return None


class FakeBot(BotClient):
    def __init__(self) -> None:
        self.command_calls: list[dict] = []
        self.callback_calls: list[dict] = []
        self.send_calls: list[dict] = []
        self.document_calls: list[dict] = []
        self.edit_calls: list[dict] = []
        self.edit_topic_calls: list[dict[str, Any]] = []
        self.delete_calls: list[dict] = []

    async def get_updates(
        self,
        offset: int | None,
        timeout_s: int = 50,
        allowed_updates: list[str] | None = None,
    ) -> list[Update] | None:
        _ = offset
        _ = timeout_s
        _ = allowed_updates
        return []

    async def get_file(self, file_id: str) -> File | None:
        _ = file_id
        return None

    async def download_file(self, file_path: str) -> bytes | None:
        _ = file_path
        return None

    async def send_message(
        self,
        chat_id: int,
        text: str,
        reply_to_message_id: int | None = None,
        disable_notification: bool | None = False,
        message_thread_id: int | None = None,
        entities: list[dict[str, Any]] | None = None,
        parse_mode: str | None = None,
        reply_markup: dict | None = None,
        *,
        replace_message_id: int | None = None,
    ) -> Message:
        self.send_calls.append(
            {
                "chat_id": chat_id,
                "text": text,
                "reply_to_message_id": reply_to_message_id,
                "disable_notification": disable_notification,
                "message_thread_id": message_thread_id,
                "entities": entities,
                "parse_mode": parse_mode,
                "reply_markup": reply_markup,
                "replace_message_id": replace_message_id,
            }
        )
        return Message(message_id=1, chat=Chat(id=chat_id, type="private"))

    async def send_document(
        self,
        chat_id: int,
        filename: str,
        content: bytes,
        reply_to_message_id: int | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = False,
        caption: str | None = None,
    ) -> Message:
        self.document_calls.append(
            {
                "chat_id": chat_id,
                "filename": filename,
                "content": content,
                "reply_to_message_id": reply_to_message_id,
                "message_thread_id": message_thread_id,
                "disable_notification": disable_notification,
                "caption": caption,
            }
        )
        return Message(message_id=2, chat=Chat(id=chat_id, type="private"))

    async def edit_message_text(
        self,
        chat_id: int,
        message_id: int,
        text: str,
        entities: list[dict[str, Any]] | None = None,
        parse_mode: str | None = None,
        reply_markup: dict | None = None,
        *,
        wait: bool = True,
    ) -> Message:
        self.edit_calls.append(
            {
                "chat_id": chat_id,
                "message_id": message_id,
                "text": text,
                "entities": entities,
                "parse_mode": parse_mode,
                "reply_markup": reply_markup,
                "wait": wait,
            }
        )
        return Message(message_id=message_id, chat=Chat(id=chat_id, type="private"))

    async def delete_message(self, chat_id: int, message_id: int) -> bool:
        self.delete_calls.append({"chat_id": chat_id, "message_id": message_id})
        return True

    async def set_my_commands(
        self,
        commands: list[dict[str, Any]],
        *,
        scope: dict[str, Any] | None = None,
        language_code: str | None = None,
    ) -> bool:
        self.command_calls.append(
            {
                "commands": commands,
                "scope": scope,
                "language_code": language_code,
            }
        )
        return True

    async def get_me(self) -> User | None:
        return User(id=1, username="bot")

    async def get_chat(self, chat_id: int) -> Chat | None:
        _ = chat_id
        return Chat(id=chat_id, type="supergroup", is_forum=True)

    async def get_chat_member(self, chat_id: int, user_id: int) -> ChatMember | None:
        _ = chat_id
        _ = user_id
        return ChatMember(status="administrator", can_manage_topics=True)

    async def create_forum_topic(self, chat_id: int, name: str) -> ForumTopic | None:
        _ = chat_id
        _ = name
        return ForumTopic(message_thread_id=1)

    async def edit_forum_topic(
        self, chat_id: int, message_thread_id: int, name: str
    ) -> bool:
        self.edit_topic_calls.append(
            {
                "chat_id": chat_id,
                "message_thread_id": message_thread_id,
                "name": name,
            }
        )
        return True

    async def close(self) -> None:
        return None

    async def answer_callback_query(
        self,
        callback_query_id: str,
        text: str | None = None,
        show_alert: bool | None = None,
    ) -> bool:
        self.callback_calls.append(
            {
                "callback_query_id": callback_query_id,
                "text": text,
                "show_alert": show_alert,
            }
        )
        return True


def make_cfg(
    transport: FakeTransport,
    runner: ScriptRunner | None = None,
    *,
    engine_id: str = DEFAULT_ENGINE_ID,
    forward_coalesce_s: float = 0.0,
    media_group_debounce_s: float = 0.0,
) -> TelegramBridgeConfig:
    if runner is None:
        runner = ScriptRunner([Return(answer="ok")], engine=engine_id)
    exec_cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    runtime = TransportRuntime(
        router=_make_router(runner),
        projects=_empty_projects(),
    )
    return TelegramBridgeConfig(
        bot=FakeBot(),
        runtime=runtime,
        chat_id=123,
        startup_msg="",
        exec_cfg=exec_cfg,
        forward_coalesce_s=forward_coalesce_s,
        media_group_debounce_s=media_group_debounce_s,
    )



from __future__ import annotations

from typing import Literal

from ..transport_runtime import TransportRuntime
from .chat_prefs import ChatPrefsStore
from .commands.parse import _parse_slash_command
from .topic_state import TopicStateStore
from .types import TelegramIncomingMessage

TriggerMode = Literal["all", "mentions"]


async def resolve_trigger_mode(
    *,
    chat_id: int,
    thread_id: int | None,
    chat_prefs: ChatPrefsStore | None,
    topic_store: TopicStateStore | None,
) -> TriggerMode:
    if topic_store is not None and thread_id is not None:
        topic_mode = await topic_store.get_trigger_mode(chat_id, thread_id)
        if topic_mode == "mentions":
            return "mentions"
    if chat_prefs is not None:
        chat_mode = await chat_prefs.get_trigger_mode(chat_id)
        if chat_mode == "mentions":
            return "mentions"
    return "all"


def should_trigger_run(
    msg: TelegramIncomingMessage,
    *,
    bot_username: str | None,
    runtime: TransportRuntime,
    command_ids: set[str],
    reserved_chat_commands: set[str],
) -> bool:
    text = msg.text or ""
    lowered = text.lower()
    if bot_username:
        needle = f"@{bot_username}"
        if needle in lowered:
            return True
    implicit_topic_reply = (
        msg.thread_id is not None and msg.reply_to_message_id == msg.thread_id
    )

    if msg.reply_to_is_bot and not implicit_topic_reply:
        return True
    if (
        bot_username
        and msg.reply_to_username
        and msg.reply_to_username.lower() == bot_username
        and not implicit_topic_reply
    ):
        return True
    command_id, _ = _parse_slash_command(text)
    if not command_id:
        return False
    if command_id in reserved_chat_commands or command_id in command_ids:
        return True
    engine_ids = {engine.lower() for engine in runtime.available_engine_ids()}
    if command_id in engine_ids:
        return True
    project_aliases = {alias.lower() for alias in runtime.project_aliases()}
    return command_id in project_aliases



from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class ActionPlan:
    reply_text: str | None
    actions: tuple[Callable[[], Awaitable[None]], ...] = ()

    async def execute(self, reply: Callable[..., Awaitable[None]]) -> None:
        for action in self.actions:
            await action()
        if self.reply_text is not None:
            await reply(text=self.reply_text)



from __future__ import annotations

import re as _re
from typing import TYPE_CHECKING

from ...context import RunContext
from ...markdown import MarkdownParts
from ...transport_runtime import TransportRuntime
from ...transport import RenderedMessage, SendOptions
from ..chat_prefs import ChatPrefsStore
from ..chat_sessions import ChatSessionStore
from ..context import (
    _format_context,
    _format_ctx_status,
    _merge_topic_context,
    _parse_project_branch_args,
    _usage_ctx_set,
    _usage_topic,
)
from ..files import split_command_args
from ..render import prepare_telegram
from ..topic_state import TopicStateStore
from ..topics import (
    _maybe_rename_topic,
    _topic_key,
    _topic_title,
    _topics_chat_project,
    _topics_command_error,
)
from ..types import TelegramIncomingMessage
from .reply import make_reply

if TYPE_CHECKING:
    from ..bridge import TelegramBridgeConfig


async def _handle_ctx_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    args_text: str,
    store: TopicStateStore,
    *,
    resolved_scope: str | None = None,
    scope_chat_ids: frozenset[int] | None = None,
) -> None:
    reply = make_reply(cfg, msg)
    error = _topics_command_error(
        cfg,
        msg.chat_id,
        resolved_scope=resolved_scope,
        scope_chat_ids=scope_chat_ids,
    )
    if error is not None:
        await reply(text=error)
        return
    chat_project = _topics_chat_project(cfg, msg.chat_id)
    tkey = _topic_key(msg, cfg, scope_chat_ids=scope_chat_ids)
    if tkey is None:
        await reply(text="this command only works inside a topic.")
        return
    tokens = split_command_args(args_text)
    action = tokens[0].lower() if tokens else "show"
    if action in {"show", ""}:
        snapshot = await store.get_thread(*tkey)
        bound = snapshot.context if snapshot is not None else None
        ambient = _merge_topic_context(chat_project=chat_project, bound=bound)
        resolved = cfg.runtime.resolve_message(
            text="",
            reply_text=msg.reply_to_text,
            chat_id=msg.chat_id,
            ambient_context=ambient,
        )
        engine = None
        model = None
        if snapshot is not None:
            engine = snapshot.default_engine
            if engine is not None:
                override = await store.get_engine_override(tkey[0], tkey[1], engine)
                if override is not None and override.model is not None:
                    model = override.model
        text = _format_ctx_status(
            cfg=cfg,
            runtime=cfg.runtime,
            bound=bound,
            resolved=resolved.context,
            context_source=resolved.context_source,
            snapshot=snapshot,
            chat_project=chat_project,
            engine=engine,
            model=model,
        )
        await reply(text=text)
        return
    if action == "set":
        rest = " ".join(tokens[1:])
        context, error = _parse_project_branch_args(
            rest,
            runtime=cfg.runtime,
            require_branch=False,
            chat_project=chat_project,
        )
        if error is not None:
            await reply(
                text=f"error:\n{error}\n{_usage_ctx_set(chat_project=chat_project)}",
            )
            return
        if context is None:
            await reply(
                text=f"error:\n{_usage_ctx_set(chat_project=chat_project)}",
            )
            return
        await store.set_context(*tkey, context)
        await _maybe_rename_topic(
            cfg,
            store,
            chat_id=tkey[0],
            thread_id=tkey[1],
            context=context,
        )
        await reply(
            text=f"topic bound to `{_format_context(cfg.runtime, context)}`",
        )
        return
    if action == "clear":
        await store.clear_context(*tkey)
        await reply(text="topic binding cleared.")
        return
    await reply(
        text="unknown `/ctx` command. use `/ctx`, `/ctx set`, or `/ctx clear`.",
    )


def _parse_chat_ctx_args(
    args_text: str,
    *,
    runtime: TransportRuntime,
    default_project: str | None,
) -> tuple[RunContext | None, str | None]:
    tokens = split_command_args(args_text)
    if not tokens:
        return None, _usage_ctx_set(chat_project=None)
    if len(tokens) > 2:
        return None, "too many arguments"
    project_token: str | None = None
    branch: str | None = None
    first = tokens[0]
    if first.startswith("@"):
        branch = first[1:] or None
    else:
        project_token = first
        if len(tokens) == 2:
            second = tokens[1]
            if not second.startswith("@"):
                return None, "branch must be prefixed with @"
            branch = second[1:] or None
    project_key: str | None = None
    if project_token is None:
        if default_project is None:
            return None, "project is required"
        project_key = default_project
    else:
        project_key = runtime.normalize_project_key(project_token)
        if project_key is None:
            return None, f"unknown project {project_token!r}"
    return RunContext(project=project_key, branch=branch), None


async def _handle_chat_ctx_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    args_text: str,
    chat_prefs: ChatPrefsStore | None,
) -> None:
    reply = make_reply(cfg, msg)
    if chat_prefs is None:
        await reply(text="chat context unavailable; config path is not set.")
        return

    tokens = split_command_args(args_text)
    action = tokens[0].lower() if tokens else "show"
    if action in {"show", ""}:
        bound = await chat_prefs.get_context(msg.chat_id)
        resolved = cfg.runtime.resolve_message(
            text="",
            reply_text=msg.reply_to_text,
            chat_id=msg.chat_id,
            ambient_context=bound,
        )
        source = resolved.context_source
        if bound is not None and resolved.context_source == "ambient":
            source = "bound"
        lines = [
            f"bound ctx: {_format_context(cfg.runtime, bound)}",
            f"resolved ctx: {_format_context(cfg.runtime, resolved.context)} (source: {source})",
        ]
        if bound is None:
            ctx_usage = (
                _usage_ctx_set(chat_project=None).removeprefix("usage: ").strip()
            )
            lines.append(f"note: no bound context — bind with {ctx_usage}")
        await reply(text="\n".join(lines))
        return
    if action == "set":
        rest = " ".join(tokens[1:])
        context, error = _parse_chat_ctx_args(
            rest,
            runtime=cfg.runtime,
            default_project=cfg.runtime.default_project,
        )
        if error is not None:
            await reply(
                text=f"error:\n{error}\n{_usage_ctx_set(chat_project=None)}",
            )
            return
        if context is None:
            await reply(text=f"error:\n{_usage_ctx_set(chat_project=None)}")
            return
        await chat_prefs.set_context(msg.chat_id, context)
        await reply(
            text=f"chat bound to `{_format_context(cfg.runtime, context)}`",
        )
        return
    if action == "clear":
        await chat_prefs.clear_context(msg.chat_id)
        await reply(text="chat context cleared.")
        return
    await reply(
        text="unknown `/ctx` command. use `/ctx`, `/ctx set`, or `/ctx clear`.",
    )


async def _handle_new_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    store: TopicStateStore,
    *,
    resolved_scope: str | None = None,
    scope_chat_ids: frozenset[int] | None = None,
) -> None:
    reply = make_reply(cfg, msg)
    error = _topics_command_error(
        cfg,
        msg.chat_id,
        resolved_scope=resolved_scope,
        scope_chat_ids=scope_chat_ids,
    )
    if error is not None:
        await reply(text=error)
        return
    tkey = _topic_key(msg, cfg, scope_chat_ids=scope_chat_ids)
    if tkey is None:
        await reply(text="this command only works inside a topic.")
        return
    await store.clear_sessions(*tkey)
    await reply(text="cleared stored sessions for this topic.")


async def _handle_chat_new_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    store: ChatSessionStore,
    session_key: tuple[int, int | None] | None,
) -> None:
    reply = make_reply(cfg, msg)
    if session_key is None:
        await reply(text="no stored sessions to clear for this chat.")
        return
    await store.clear_sessions(session_key[0], session_key[1])
    if msg.chat_type == "private":
        text = "cleared stored sessions for this chat."
    else:
        text = "cleared stored sessions for you in this chat."
    await reply(text=text)


async def _handle_topic_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    args_text: str,
    store: TopicStateStore,
    *,
    resolved_scope: str | None = None,
    scope_chat_ids: frozenset[int] | None = None,
) -> None:
    reply = make_reply(cfg, msg)
    error = _topics_command_error(
        cfg,
        msg.chat_id,
        resolved_scope=resolved_scope,
        scope_chat_ids=scope_chat_ids,
    )
    if error is not None:
        await reply(text=error)
        return
    chat_project = _topics_chat_project(cfg, msg.chat_id)
    context, error = _parse_project_branch_args(
        args_text,
        runtime=cfg.runtime,
        require_branch=True,
        chat_project=chat_project,
    )
    if error is not None or context is None:
        usage = _usage_topic(chat_project=chat_project)
        text = f"error:\n{error}\n{usage}" if error else usage
        await reply(text=text)
        return
    title = _topic_title(runtime=cfg.runtime, context=context)
    existing = await store.find_thread_for_context(msg.chat_id, context)
    stale_thread_id: int | None = None
    if existing is not None:
        updated = await cfg.bot.edit_forum_topic(
            chat_id=msg.chat_id,
            message_thread_id=existing,
            name=title,
        )
        if updated:
            await reply(
                text=f"topic already exists for {_format_context(cfg.runtime, context)} "
                "in this chat.",
            )
            return
        stale_thread_id = existing
    created = await cfg.bot.create_forum_topic(msg.chat_id, title)
    if created is None:
        await reply(text="failed to create topic.")
        return
    thread_id = created.message_thread_id
    if stale_thread_id is not None:
        await store.delete_thread(msg.chat_id, stale_thread_id)
    await store.set_context(
        msg.chat_id,
        thread_id,
        context,
        topic_title=title,
    )
    await reply(text=f"created topic `{title}`.")
    bound_text = f"topic bound to `{_format_context(cfg.runtime, context)}`"
    rendered_text, entities = prepare_telegram(MarkdownParts(header=bound_text))
    await cfg.exec_cfg.transport.send(
        channel_id=msg.chat_id,
        message=RenderedMessage(text=rendered_text, extra={"entities": entities}),
        options=SendOptions(thread_id=thread_id),
    )


_FORK_TITLE_PATTERN = _re.compile(r"^(.+) \(fork #(\d+)\)$")


def _get_forked_title(title: str | None) -> str:
    if title is None:
        return "topic (fork #1)"
    match = _FORK_TITLE_PATTERN.match(title)
    if match:
        base = match.group(1)
        num = int(match.group(2))
        return f"{base} (fork #{num + 1})"
    return f"{title} (fork #1)"


async def _handle_fork_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    store: TopicStateStore,
    *,
    resolved_scope: str | None = None,
    scope_chat_ids: frozenset[int] | None = None,
) -> None:
    from ...model import ResumeToken

    reply = make_reply(cfg, msg)
    error = _topics_command_error(
        cfg,
        msg.chat_id,
        resolved_scope=resolved_scope,
        scope_chat_ids=scope_chat_ids,
    )
    if error is not None:
        await reply(text=error)
        return
    tkey = _topic_key(msg, cfg, scope_chat_ids=scope_chat_ids)
    if tkey is None:
        await reply(text="this command only works inside a topic.")
        return

    snapshot = await store.get_thread(*tkey)
    current_title = snapshot.topic_title if snapshot else None
    forked_title = _get_forked_title(current_title)

    created = await cfg.bot.create_forum_topic(msg.chat_id, forked_title)
    if created is None:
        await reply(text="failed to create forked topic.")
        return

    new_thread_id = created.message_thread_id

    if snapshot and snapshot.context:
        await store.set_context(
            msg.chat_id,
            new_thread_id,
            snapshot.context,
            topic_title=forked_title,
        )

    if snapshot and snapshot.sessions:
        for engine, resume_value in snapshot.sessions.items():
            token = ResumeToken(engine=engine, value=resume_value)
            await store.set_session_resume(msg.chat_id, new_thread_id, token)

    await reply(text=f"forked to new topic `{forked_title}`.")

    context_label = _format_context(cfg.runtime, snapshot.context) if snapshot and snapshot.context else "none"
    sessions_label = ", ".join(sorted(snapshot.sessions.keys())) if snapshot and snapshot.sessions else "none"
    welcome_text = (
        f"forked from previous topic\n\n"
        f"context: `{context_label}`\n"
        f"sessions: {sessions_label}\n\n"
        "continue your conversation here!"
    )
    rendered_text, entities = prepare_telegram(MarkdownParts(header=welcome_text))
    await cfg.exec_cfg.transport.send(
        channel_id=msg.chat_id,
        message=RenderedMessage(text=rendered_text, extra={"entities": entities}),
        options=SendOptions(thread_id=new_thread_id),
    )



# /// script
# requires-python = ">=3.14"
# dependencies = [
#     "mistune>=3.2.0",
#     "requests>=2.32.5",
#     "sulguk>=0.11.1",
# ]
# ///
from __future__ import annotations

import json
import os
from pathlib import Path
import re

import mistune
import requests
import sulguk

repo = os.environ["REPO"]
bot_token = os.environ["TELEGRAM_BOT_TOKEN"]
chat_id = os.environ["TELEGRAM_CHAT_ID"]

event_path = Path(os.environ["GITHUB_EVENT_PATH"])
event = json.loads(event_path.read_text(encoding="utf-8"))

ref = event.get("ref") or ""
branch = ref.removeprefix("refs/heads/") if ref.startswith("refs/heads/") else ref
commits = list(event.get("commits") or [])
head_commit = event.get("head_commit")
if not commits and head_commit:
    commits = [head_commit]

PULL_RE = re.compile(rf"(https://github.com/{repo}/pull/(\d+))")
PULL_NUM_RE = re.compile(r"\(#(\d+)\)")


def _short_sha(value: str) -> str:
    return value[:7] if value else ""


def _commit_line(commit: dict[str, object]) -> str:
    full_sha = str(commit.get("id") or "")
    sha = _short_sha(full_sha)
    message = str(commit.get("message") or "").splitlines()[0].strip()
    message = PULL_RE.sub(r"[#\2](\1)", message)
    message = PULL_NUM_RE.sub(
        rf"([#\1](https://github.com/{repo}/pull/\1))", message
    )
    url = f"https://github.com/{repo}/commit/{full_sha or sha}"
    return f"- [{sha}]({url}) {message}"


lines: list[str] = []
if commits:
    max_commits = 10
    lines.extend(_commit_line(commit) for commit in commits[:max_commits])
    if len(commits) > max_commits:
        lines.append(f"- ...and {len(commits) - max_commits} more")

header = f"push **{repo} {branch}**".strip()
parts = [header]
if lines:
    parts.append("\n".join(lines))

message = "\n\n".join(part for part in parts if part)

html = mistune.html(message)
rendered = sulguk.transform_html(html)

payload = {
    "chat_id": chat_id,
    "text": rendered.text,
    "entities": rendered.entities,
    "link_preview_options": {"is_disabled": True},
}
resp = requests.post(
    f"https://api.telegram.org/bot{bot_token}/sendMessage", json=payload
)
resp.raise_for_status()
print(f"sent to {chat_id}")



from pathlib import Path

import pytest
from typer.testing import CliRunner

from yee88 import cli
from yee88.config import ConfigError, read_config
from yee88.ids import RESERVED_CHAT_COMMANDS
from yee88.settings import TakopiSettings


def _base_config() -> dict:
    return {"transports": {"telegram": {"bot_token": "token", "chat_id": 123}}}


def test_parse_projects_rejects_engine_alias() -> None:
    config = {**_base_config(), "projects": {"codex": {"path": "/tmp/repo"}}}
    with pytest.raises(ConfigError, match="aliases must not match engine ids"):
        settings = TakopiSettings.model_validate(config)
        settings.to_projects_config(
            config_path=Path("yee88.toml"),
            engine_ids=["codex"],
            reserved=RESERVED_CHAT_COMMANDS,
        )


def test_parse_projects_default_project_must_exist() -> None:
    config = {**_base_config(), "default_project": "z80", "projects": {}}
    with pytest.raises(ConfigError, match="default_project"):
        settings = TakopiSettings.model_validate(config)
        settings.to_projects_config(
            config_path=Path("yee88.toml"),
            engine_ids=["codex"],
            reserved=RESERVED_CHAT_COMMANDS,
        )


def test_init_writes_project(monkeypatch, tmp_path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n\n[transports.telegram]\n'
        'bot_token = "token"\nchat_id = 123\n',
        encoding="utf-8",
    )
    monkeypatch.setattr("yee88.config.HOME_CONFIG_PATH", config_path)
    monkeypatch.setattr(cli, "resolve_default_base", lambda _: "main")
    monkeypatch.setattr(cli, "_load_settings_optional", lambda: (None, None))

    repo_path = tmp_path / "repo"
    repo_path.mkdir()
    monkeypatch.chdir(repo_path)

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["init", "z80"])
    assert result.exit_code == 0

    saved = config_path.read_text(encoding="utf-8")
    assert "[projects.z80]" in saved
    assert 'worktrees_dir = ".worktrees"' in saved
    assert 'default_engine = "codex"' in saved
    assert 'worktree_base = "main"' in saved


def test_init_migrates_legacy_config(monkeypatch, tmp_path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text('bot_token = "token"\nchat_id = 123\n', encoding="utf-8")
    monkeypatch.setattr("yee88.config.HOME_CONFIG_PATH", config_path)
    monkeypatch.setattr(cli, "resolve_default_base", lambda _: "main")
    monkeypatch.setattr(cli, "_load_settings_optional", lambda: (None, None))

    repo_path = tmp_path / "repo"
    repo_path.mkdir()
    monkeypatch.chdir(repo_path)

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["init", "z80"])
    assert result.exit_code == 0

    raw = read_config(config_path)
    assert "bot_token" not in raw
    assert "chat_id" not in raw
    assert raw["transport"] == "telegram"
    assert raw["transports"]["telegram"]["bot_token"] == "token"
    assert raw["transports"]["telegram"]["chat_id"] == 123
    assert "z80" in raw.get("projects", {})


def test_projects_default_engine_unknown() -> None:
    config = {
        **_base_config(),
        "projects": {"z80": {"path": "/tmp/repo", "default_engine": "nope"}},
    }
    settings = TakopiSettings.model_validate(config)
    with pytest.raises(ConfigError, match=r"projects\.z80\.default_engine"):
        settings.to_projects_config(
            config_path=Path("yee88.toml"),
            engine_ids=["codex"],
            reserved=RESERVED_CHAT_COMMANDS,
        )


def test_projects_chat_id_cannot_match_transport_chat_id() -> None:
    config = {
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        "projects": {"z80": {"path": "/tmp/repo", "chat_id": 123}},
    }
    settings = TakopiSettings.model_validate(config)
    with pytest.raises(ConfigError, match="chat_id"):
        settings.to_projects_config(
            config_path=Path("yee88.toml"),
            engine_ids=["codex"],
            reserved=RESERVED_CHAT_COMMANDS,
        )


def test_projects_chat_id_must_be_unique() -> None:
    config = {
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        "projects": {
            "a": {"path": "/tmp/a", "chat_id": -10},
            "b": {"path": "/tmp/b", "chat_id": -10},
        },
    }
    settings = TakopiSettings.model_validate(config)
    with pytest.raises(ConfigError, match="chat_id"):
        settings.to_projects_config(
            config_path=Path("yee88.toml"),
            engine_ids=["codex"],
            reserved=RESERVED_CHAT_COMMANDS,
        )


def test_projects_relative_path_resolves(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    settings = TakopiSettings.model_validate(
        {**_base_config(), "projects": {"z80": {"path": "repo"}}}
    )
    projects = settings.to_projects_config(
        config_path=config_path,
        engine_ids=["codex"],
        reserved=RESERVED_CHAT_COMMANDS,
    )
    assert projects.projects["z80"].path == config_path.parent / "repo"



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any


def atomic_write_json(
    path: Path,
    payload: Any,
    *,
    indent: int = 2,
    sort_keys: bool = True,
) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = path.with_suffix(f"{path.suffix}.tmp")
    with open(tmp_path, "w", encoding="utf-8") as handle:
        json.dump(payload, handle, indent=indent, sort_keys=sort_keys)
        handle.write("\n")
    os.replace(tmp_path, path)
